// >>>>> do_check.cc


#include "check_cfg.h"

char** abc;


void compare_classes(char* name1, mmi_result& r1, char* name2, mmi_result& r2, char** abc)
{
  int ok=1;  
  int A=r1.A, C=r1.C;
  if(A!=r2.A || C!=r2.C){
    printf("FATAL: %s A=%d C=%d\n"
           "       %s A=%d C=%d\n", name1,r1.A,r1.C, name2,r2.A,r2.C);
    return;
  }
  
  FF ls1=0,ls2=0;
  for(int k=0; k<C-1; k++) {ls1+=r1.loss[k];ls2+=r2.loss[k];}

  for(int k=A; k<A+C-1; k++){  //normalize order
    if(r1.forest[k].l > r1.forest[k].r) swap(r1.forest[k].l, r1.forest[k].r);
    if(r2.forest[k].l > r2.forest[k].r) swap(r2.forest[k].l, r2.forest[k].r);
  }
  
  int eps_mult=15+5*C;
  
  if(!is_near(ls1,ls2,eps_mult)){
    ok=0;
    printf("%s loss sum=%25.23f\n", name1, D(ls1));
    printf("%s loss sum=%25.23f\n", name2, D(ls2));
  }
  
  if(!is_near(r1.I, r2.I,eps_mult)){
    ok=0;
    printf("%s total MI=%25.23f\n", name1, D(r1.I));
    printf("%s total MI=%25.23f\n", name2, D(r2.I));
  }
  
  for(int k=A; k<A+C-1; k++){
    int diff=r1.forest[k].l != r2.forest[k].l || r1.forest[k].r != r2.forest[k].r;
    if(diff || !is_near(r1.loss[k-A],r2.loss[k-A],eps_mult)){
      ok=0;
      if(diff) printf("!!");
      else printf("  ");
      printf("%5d: %s %25.23f ", k-A, name1, D(r1.loss[k-A]));
      int w=r1.forest[k].l;
      if(w<A) printf("%15s ", abc[w]);
      else printf("%13d  ", w-A);
      w=r1.forest[k].r;
      if(w<A) printf("%15s ", abc[w]);
      else printf("%13d  ", w-A);
      printf("\n");
      printf("         %s %25.23f ", name2, D(r2.loss[k-A]));
      w=r2.forest[k].l;
      if(w<A) printf("%15s ", abc[w]);
      else printf("%13d  ", w-A);
      w=r2.forest[k].r;
      if(w<A) printf("%15s ", abc[w]);
      else printf("%13d  ", w-A);
      printf("\n\n");
    }
  }
  if(ok) printf("OK, no significant differences found ;-)\n");
}


int main(int argc, char**argv)
{
  stdmsg=stdout;
  u8* bfr=0; int bfr_len=0;
  int A UNUSED; 
  A=load_dictionary(abc, argv[1]);
  int L=bin_load_file(bfr, bfr_len, argv[1], argv[2])/4;
  int* w=(int*)bfr;
   
  int threshold=atoi(argv[3]);
  mmi_result r1;
  FIRST(w, w+1, L-1, threshold, r1);
  mmi_result r2;
  SECOND(w, w+1, L-1, threshold, r2);
  printf("RESULTS:\n\n");
  print_tree(stdout,r2, abc);
  printf("\n\nDIFFERENCES BETWEEN IMPLEMENTATIONS:\n\n");
  compare_classes(FIRST_NAME,r1, SECOND_NAME, r2, abc);
  return 0;
}


