// >>>>> classes.cc

#define ERROR_MSG_PREFIX "mmi_classes: "
#include "base.cc"


void mark_subtree(int* map, mmi_result &r, int k, int mark)
{
  map[k]=mark;
  int left=r.forest[k].l;
  int right=r.forest[k].r;
  if(left>0)  mark_subtree(map, r, left, mark);
  if(right>0) mark_subtree(map, r, right, mark);
}

void unmark_tree(int* map, mmi_result &r, int k) 
{
  map[k]=-1;
  int left=r.forest[k].l;
  int right=r.forest[k].r;
  if(left>0) unmark_tree(map,r,left);
  if(right>0) unmark_tree(map,r,right);
}

void print_class(mmi_result& r, int* map, char** abc, int* hist, int k)
{
  const int width=79;
  D sum=0;
  for(int l=1; l<r.A; l++) if(map[l]==k) sum+=hist[l];
  printf("class %d: %6.3f%% \n", k, sum/hist[0]*100.0);
  int xpos=0;
  for(int l=1; l<r.A; l++) if(map[l]==k){
    if(xpos+int(strlen(abc[l]))+2>width){
      xpos=0; printf("\n");
    }
    xpos+=printf("  %s", abc[l]);
  }
  printf("\n\n");
}

int main(int argc, char** argv)
{
  if(argc<3){
    printf(
"tree cutter / class former                             mmi_clustering v"VERSION"\n"
    "Usage: classes x.tree NR\n"
    "  where NR is the desired number of classes. Will make file x.blocks\n"
    );
    return 0;
  }
  mmi_result r;
  char dirname[2048];
  char** abc;
  load_tree(argv[1], r, dirname, abc);  
  
  int NR=atoi(argv[2]);
  if(NR<=1) error("NR must be at least 2, but it was %d", NR);
  if(NR>r.C+1) error("Cannot have more than C+1=%d classes (NR=%d)", r.C+1, NR);

  int map[2*r.A]; fill(map, 0, 2*r.A); //default class
  unmark_tree(map, r, r.A+r.C-2);
  
  int mark=1;
  for(int k=r.A+r.C-NR; k>0; k--) if(map[k]<0) mark_subtree(map, r, k, mark++);
  
  int up[2*r.A]; fill(up, 0, 2*r.A);
  for(int k=r.A+r.C-2; k>=r.A; k--)
    up[r.forest[k].l]=up[r.forest[k].r]=k;

//  if(sigletons_sw){
    for(int k=1; k<r.A; k++) if(map[k]==0) map[k]=mark++;
//  }
  char fn[strlen(argv[1])+20]; strcpy(fn,argv[1]);
  char* dot=strrchr(fn,'.'); if(!dot) dot=fn+strlen(fn);
  strcpy(dot,".blocks");
  save_eqmap(fn, abc, map, 0, r.A-1);
  
  return 0;
}

