// >>>>> mmi_clustering_1.cc

#include "base.cc"


/* Better implemantation than mmi_clustering_0.cc. May not return some
 * memory it allocats, but should return right results. Intended to be
 * compared against mmi_clustering_0.cc and then serve as a baseline against
 * which we will verify mmi_clustering_2.cc (we cannot use simple
 * mmi_clustering_0.cc except on really small inputs).
 */

STAT(i64 h_collision=0; i64 h_access=0;)

struct entry{
  i32 key, data;
};

struct hash{
  entry *a;
  int occupied;
  int L;
  
  void init(int n){
    DBG(if(n==0) error("hash.init():n=0!!!\n");)
    occupied=0; L=n; alloc(a,n); fill(a, ((entry){0,0}), n);
  }
  void done() { dispose(a); L=0; occupied=0; }
  void init() {a=0; L=0; occupied=0;}
  hash() { init(); }
  ~hash() { done(); }
  entry* hash_qry(i32 key){
    DBG(if(!a) {error("hash_qry: use of empty hash!!");})
    STAT(h_access++;)
    i32 idx=key%L;
    i32 k=a[idx].key;
    if(k==key || k==0) return &a[idx];
    i32 idx2=(key^0x7fffffff)%L;
    idx=idx2;
    i32 idxf=-1;
    do{
      STAT(h_collision++;)
      k=a[idx].key;
      if(k==key || k==0) return &a[idx];
      if(k<0 && idxf<0) idxf=idx;
      idx=(idx+1)%L;
    }while(idx!=idx2);
    if (idxf>=0) return &a[idxf];
    return 0;
  }
  i32 hash_data(i32 key){   //to be used for data reading, non-present entry returns 0    
    i32 idx=key%L;          //equivalent to {p=hash_qry(key); return p?(p->key>0?p->data:0):0;}
    DBG(if(!a) {error("hash_qry: use of emtpy hash!!");})
    STAT(h_access++;)
    i32 k=a[idx].key;
    if(k==key) return a[idx].data;
    if(k==0) return 0;
    i32 idx2=(key^0x7fffffff)%L;
    idx=idx2;
    do{
      STAT(h_collision++;)
      k=a[idx].key;
      if(k==key) return a[idx].data;
      if(k==0) return 0;
      idx=(idx+1)%L;
    }while(idx!=idx2);
    return 0;
  } 
  void hash_ins(entry* he, i32 key, i32 data){occupied++; he->key=key; he->data=data;}
  void hash_del(entry* he) {occupied--; he->key=-1;}
};

STAT(i64 h_collision_build1; i64 h_access_build1; int h_size1;)
STAT(i64 h_collision_build2; i64 h_access_build2; int h_size2;)

struct ctab{
  hash* c;
  int A,C,N;
  int B;      //number of unique bigrams
  i32* c_map;
  i32* alias;
  i32* epoch_guard;
  i32 epoch_ct;
  i32** x_walk;
  i32* x_walk_size;
  i32* c0, *c1;
  F* Rc0, *Rc1;
  F I;

  i32 live(i32 y){
    i32 z=y; 
    if(alias[z]!=z){
      do{z=alias[z];}while(alias[z]!=z);
      do{i32 y1=alias[y]; alias[y]=z; y=y1;} while(alias[y]!=y);
    }
    return z;
  }   
  
  ~ctab() {free(c); free(c_map); for(int k=0; k<C; k++) dispose(x_walk[k]); free(x_walk);}

  void build_tab(i32* w0, i32* w1, int* ysize){
    for(int k=0; k<N; k++){
      i32 y=w0[k], x=w1[k];
      if(c[y].a==0) c[y].init(ysize[y]);
      entry* p=c[y].hash_qry(x);
      DBG(if(!p) error("build_tab: Internal error, out of space, y=%d,x=%d\n", y,x);)
      if (p->key<=0) c[y].hash_ins(p,x,i32(0));
      p->data++;
    }
  }
  
  STAT(
  int sum_size(int* size, int L){
    int acc=0;
    for(int k=0; k<L; k++) acc+=size[k];
    return acc;
  })

  ctab(i32* w0, i32* w1, i32 NN, i32 T){
    int max_w=0; N=NN;
    for(int k=0; k<N; k++) if(max_w<w0[k]) max_w=w0[k];
    for(int k=0; k<N; k++) if(max_w<w1[k]) max_w=w1[k];
    A=max_w+1; //we need space 1:max_w, so we allocate 0:max_w=[A] where [0] will never be used
    alloc(c,A); for(int a=0; a<A; a++) c[a].init(); 
    int cw0[A], cw1[A];
    fill(cw0,i32(0),A); for(int k=0; k<N; k++) cw0[w0[k]]++;
    fill(cw1,i32(0),A); for(int k=0; k<N; k++) cw1[w1[k]]++;

    alloc(epoch_guard, A); epoch_ct=0;
    alloc(alias, A); for(int a=0; a<A; a++) alias[a]=a;

    STAT(h_collision=0; h_access=0;)
    build_tab(w0,w1,cw0);  //first build of ctab is only used to compute space demands
    STAT(h_collision_build1=h_collision; h_access_build1=h_access; h_size1=sum_size(cw0, A);)    
    {
      int ct[A]; for(int a=0; a<A; a++) ct[a]=iround((1+oversize)*c[a].occupied);
      for(int a=0; a<A; a++) c[a].done(); //delete the table now
      STAT(h_collision=0; h_access=0;)
      build_tab(w0,w1,ct);                //and build the right-sized one
      STAT(h_collision_build2=h_collision; h_access_build2=h_access; h_size2=sum_size(ct, A);)
      B=0; for(int a=1; a<A; a++) B+=c[a].occupied;
    }
    STAT(h_collision=0; h_access=0;)

    int inv_c_map[A]; fill(inv_c_map, -1, A);
    C=0;
    for(int a=1; a<A; a++) if(cw0[a]+cw1[a]>=2*T) C++;
    alloc(c_map,C);    
    for(int a=1,i=0; a<A; a++) if(cw0[a]+cw1[a]>=2*T) {c_map[i]=a; inv_c_map[a]=i; i++;}

    alloc(x_walk,C); alloc(x_walk_size, C);
    {
      int ct[C]; fill(ct,0,C);
      for(int y=1; y<A; y++) if (c[y].a) for(int p=0; p<c[y].L; p++){
        int x=c[y].a[p].key;
        if(x>0 && inv_c_map[x]>=0) ct[inv_c_map[x]]++;
      }
      for(int k=0; k<C; k++){
        alloc(x_walk[k],ct[k]);
        x_walk_size[k]=ct[k];
      }
    }
    int last[C]; fill(last,0,C);
    I=0;
    for(int y=1; y<A; y++) if (c[y].a){
      for(int p=0; p<c[y].L; p++) {
        entry* e=&(c[y].a[p]);
        i32 x=e->key;
        if(x>0){
          i32 cyx=e->data;
          I-=cyx*ulog2( F(cw0[y])*cw1[x] / (F(N)*cyx) );          
          int xx=inv_c_map[x];
          if(xx>=0) x_walk[xx][last[xx]++]=y;
        }
      }
    }
    I/=N;
    alloc(c0,C); alloc(c1,C); alloc(Rc0,C); alloc(Rc1,C);    
    for(int k=0; k<C; k++){
      c0[k]=cw0[c_map[k]]; Rc0[k]=R(c0[k]);
      c1[k]=cw1[c_map[k]]; Rc1[k]=R(c1[k]);
    }
  }

  void merge_columns(i32 ll, i32 rr){
    if(epoch_ct==0) fill(epoch_guard,i32(0),A);
    epoch_ct++;
    DBG(if(ll>=rr) error("merge_columns(): %d<%d violated\n", ll, rr);)
    int l=c_map[ll], r=c_map[rr];
    int bfr[x_walk_size[rr]], bp=0;
    for(int yp=0; yp<x_walk_size[rr]; yp++){
      i32 y=x_walk[rr][yp];
      y=live(y);
      if(epoch_guard[y]==epoch_ct) continue;
      epoch_guard[y]=epoch_ct;          /*c[y]!=NULL*/  DBG(if(!c[y].a) error("merge_columns: c[y].a==NULL");)
      entry* hr=c[y].hash_qry(r);      /*always finds*/ DBG(if(!hr || hr->key<=0) error("merge_columns: 'always finds' violated y=%d (was %d) x=%d key=%d\n",y, x_walk[rr][yp], r, hr? hr->key:-2);)
      i32 d=hr->data;
      c[y].hash_del(hr);
      entry* hl=c[y].hash_qry(l);   /*always non-null*/ DBG(if(!hl) error("merge_columns: always non-null violated");)
      if(hl->key<=0) {c[y].hash_ins(hl,l,d); bfr[bp++]=y;}
      else {hl->data+=d;}
    }
    dispose(x_walk[rr]); x_walk_size[rr]=0;
    if(bp>0){
      i32* new_x_walk; alloc(new_x_walk, x_walk_size[ll]+bp);
      copy(new_x_walk, x_walk[ll], x_walk_size[ll]);
      copy(new_x_walk+x_walk_size[ll], bfr, bp);
      x_walk_size[ll]+=bp;
      free(x_walk[ll]);
      x_walk[ll]=new_x_walk;
    }
  }
  
  void merge_rows(i32 ll, i32 rr){
    DBG(if(ll>=rr) error("merge_columns(): %d<%d violated\n", ll, rr);)
    int l=c_map[ll], r=c_map[rr];
    if(!c[l].a)
      c[l]=c[r];  /* move constructor */
    else if(c[r].a) {
      hash* lh=&(c[l]), *rh=&(c[r]);
      int intersection=0;  /* compute intersection size */
      if(lh->occupied<rh->occupied) swap(lh,rh);    
      for(int k=0; k<rh->L; k++) if(rh->a[k].key>0){
        entry *er=lh->hash_qry(rh->a[k].key);
        if(er && er->key>0) intersection++;
      }
      int s=0, S=lh->occupied+rh->occupied;
      entry bfr[S];        /* save values */
      for(int k=0; k<lh->L; k++) if(lh->a[k].key>0) bfr[s++]=lh->a[k];
      for(int k=0; k<rh->L; k++) if(rh->a[k].key>0) bfr[s++]=rh->a[k];
      lh->done(); rh->done();    
      lh=&(c[l]);
      lh->init(iround((1+oversize)*(S-intersection)));   /* create the merged row */ 
      for(s=0; s<S; s++){
        entry* e=lh->hash_qry(bfr[s].key);   DBG(if(!e) error("merge_rows: e!=null violated");)
        if (e->key<=0) lh->hash_ins(e,bfr[s].key,bfr[s].data);
        else e->data+=bfr[s].data;
      }      
    }
  }
  
  void merge(i32 ll, i32 rr)
  {
    merge_columns(ll,rr);
    merge_rows(ll,rr);
    c0[ll]+=c0[rr]; c1[ll]+=c1[rr];
    Rc0[ll]=R(c0[ll]); Rc1[ll]=R(c1[ll]);
    alias[c_map[rr]]=c_map[ll];
    c_map[rr]=0;
  }

  inline i32 cd(i32 y, i32 x) {
    if(!c[y].a) return 0;
    else return c[y].hash_data(x);
  }

  inline F Phi(i32 a, i32 b, i32 c, i32 d) {return R(a+b)+R(c+d)+R(a+c)+R(b+d)-R(a)-R(b)-R(c)-R(d)-R(a+b+c+d);}

  F I_correction(i32 aa, i32 bb, i32 ll, i32 rr) {
    i32 a=c_map[aa], b=c_map[bb], l=c_map[ll], r=c_map[rr];
    return (Phi(cd(a,l),cd(b,l),cd(a,r),cd(b,r)) + Phi(cd(l,a),cd(l,b),cd(r,a),cd(r,b)))/N;
  }
  
  F I_loss(i32 l, i32 r)
  {
    if(epoch_ct==0) fill(epoch_guard,i32(0),A);
    epoch_ct++;
    int ml=c_map[l], mr=c_map[r];
    i32 cll=cd(ml,ml), clr=cd(ml,mr), crl=cd(mr,ml), crr=cd(mr,mr);
    F base = R(c0[l]+c0[r]) - Rc0[l] - Rc0[r] + R(c1[l]+c1[r]) - Rc1[l] - Rc1[r]
           + R(cll+clr) + R(crl+crr) + R(cll+crl) + R(clr+crr)
           - R(cll)-R(clr)-R(crl)-R(crr)-R(cll+clr+crl+crr);
    F acc=0;
    if(c[ml].occupied < c[mr].occupied) {swap(ml,mr);swap(l,r);}
    for(int p=c[mr].L-1; p>=0; p--){
      i32 x=c[mr].a[p].key;
      if(x>0){
        i32 clx=cd(ml,x);
        if(clx>0){              /* update is non-zero in intersection only */
          F crx=c[mr].a[p].data;
          F alfa=clx/F(clx+crx), beta=1-alfa; /* beta=crx/F(clx+crx) */
          acc+=clx*ulog2(alfa) + crx*ulog2(beta);  
        }
      }
    }    

    if(x_walk_size[l]<x_walk_size[r]) {swap(l,r);swap(ml,mr);}
    int q=0;
    for(int p=0; p<x_walk_size[r]; p++){
      int y=live(x_walk[r][p]);
      y=live(y);
      if(epoch_guard[y]==epoch_ct) continue;
      epoch_guard[y]=epoch_ct;
      x_walk[r][q++]=y;      
      int cyl=cd(y,ml);
      if(cyl>0){
        int cyr=cd(y,mr);
        F alfa=cyl/F(cyl+cyr), beta=1-alfa;
        acc+=cyl*ulog2(alfa) + cyr*ulog2(beta); 
      }
    }
    x_walk_size[r]=q;
//printf("%d %d base=%g acc=%g tot=%g\n", l, r, D(base/N), D(acc/N), D((base+acc)/N));
    return (base+acc)/N;
  }
};
static inline int yx2i(int y, int x) { return y+x*(x-1)/2; }  /* y<x */

void mmi_clustering_1(i32* w0, i32* w1, int N, int T, mmi_result &ret)
{
  STAT(int start_time=time(0);)
  VERB(printf("Building table for data of length %d with threshold %d\n", N, T);)
  ctab c(w0,w1,N,T);
  VERB(printf(" ... done A=%d B=%d C=%d\n", c.A, c.B, c.C);)
  tree t000; t000.l=0; t000.r=0; t000.level=0;
  alloc(ret.forest, c.A+c.C-1); fill(ret.forest, t000, c.A);
  alloc(ret.loss, c.C-1);
  ret.I=c.I; ret.A=c.A; ret.B=c.B; ret.C=c.C;
  int out_map[c.C]; copy(out_map, c.c_map, c.C);
  FF loss[(c.C-1)*c.C/2];  /* to be indexed by yx2i */
  for(int x=1; x<c.C; x++) for(int y=0; y<x; y++) loss[yx2i(y,x)]=c.I_loss(y,x);
  VERB(printf("initial loss table computed, starting iterations...\n"));
  for(int n=c.A; n<c.A+c.C-1; n++){
    int l=0,r=1;
    FF min_loss=F_max(FF(1));
    for(int x=1; x<c.C; x++) if(c.c_map[x]>0) for(int y=0; y<x; y++) if(c.c_map[y]>0 && min_loss>loss[yx2i(y,x)]){
      min_loss=loss[yx2i(y,x)];
      l=y; r=x;
    }
//printf(" MERGING %d %d\n", c.c_map[l], c.c_map[r]);
    for(int a=0; a<c.C; a++) if(a!=l && a!=r && c.c_map[a]>0) 
      for(int b=a+1; b<c.C; b++) if(b!=l && b!=r && c.c_map[b]>0) 
        loss[yx2i(a,b)]+=c.I_correction(l,r,a,b);
    c.merge(l,r);
//for(int y=1; y<c.A; y++) for(int x=1; x<c.A; x++) printf(" (%d %d) %d\n", y, x, c.cd(y,x));
    ret.forest[n].l=out_map[l]; ret.forest[n].r=out_map[r]; ret.loss[n-c.A]=min_loss;
    for(int y=0; y<c.C; y++) if(y!=l && c.c_map[y]>0){
      int p=y, q=l;
      if (p>q) swap(p,q);
      loss[yx2i(p,q)]=c.I_loss(p,q);
    }
//for(int x=1; x<c.C; x++) if(c.c_map[x]>0) for(int y=0; y<x; y++) if(c.c_map[y]>0)
//printf("x=%3d y=%3d  %19.16f \n", x,y, D(loss[yx2i(y,x)]));
    
    out_map[l]=n;
  }
  VERB(printf("mmi_clustering_1(): Finished in %d seconds. \n"
               " hash table1 insert count = %Ld, collisions per insert = %4.2f\n"
               "           allocated entry count = %d,  size=%4.2f MB \n"
               " hash table2 insert count = %Ld, collisions per insert = %4.2f\n"
               "           allocated entry count = %d,  size=%4.2f MB \n"
               " hash table2 access count = %Ld, collisions per access = %4.2f\n",
               int(time(0))-start_time,
               h_access_build1, D(h_collision_build1)/h_access_build1,
               h_size1, D(h_size1*sizeof(entry))/1024/1024,               
               h_access_build2, D(h_collision_build2)/h_access_build2,
               h_size2, D(h_size2*sizeof(entry))/1024/1024,
               h_access, D(h_collision)/h_access);)
}

