// >>>>> mmi_clustering_3.cc

/* void mmi_clustering_3(i32* w0, i32* w1, int N, int T, mmi_result &ret)
 *
 * classes are being made from w1, where w0 is predicted according to model
 *  k (P(w0[k]|C(w0[k])) * P(C(w0[k])|C(w1[k])))
 */


#include "base.cc"

STAT(i64 h_collision__3=0; i64 h_access__3=0; i64 h_rehash__3=0; i64 h_alloced_entries__3=0;)

struct entry__3{
  i32 key, data;  //valid keys range from 1 to (1<<31)-2 (inclusively). 0 is used to mark free space
};                //and negative values mark "deleted" entries. Value (1<<31)-1 is a sentinel which prevents writing into the empty hash

static entry__3 void_entry__3={i32(u32(1<<31)-1),-1};  //this is used as a "dummy" entry for zero-length hashes

struct hash__3{   /* NOTE: It is possible to have a zero-legth hash. In that case you must not */
  entry__3 *a;    /* write any data into it. It is safe not to call `done()' or destructor on empty hashes. */
  int occupied;   /* Empty hash can be recognized by L==-1 */
  int L;          /* L>0          ... length of the allocated space
                   * L==-1        ... the length of the hash is 0 (implies a=&void_entry__3, occupied=0)
                   * L==0 || L<-1 ... never happens
                   */  
  void init(F oversize, int n){     //note that it can be called with n==0
    if(n>2) n=iround((1+oversize)*n);
    else if(n==0) {init(); return;}
    occupied=0; L=n; alloc(a,n); fill(a, ((entry__3){0,0}), n);
    STAT(h_alloced_entries__3+=n;)
  }
  void init() {   //init() creates empty (but valid) hash
    L=-1;         //so that key%L would work (key%-1==0 for any key)
    occupied=0;
    a=&void_entry__3;
  }
  void done() { if(L>0) {free(a); STAT(h_alloced_entries__3-=L;) init();} }
  hash__3() { init(); }
  inline void move(hash__3& from) {a=from.a; occupied=from.occupied; L=from.L; from.init();}  /* move-constructor */
  ~hash__3() { done(); }
  entry__3* hash_qry(i32 key){
    DBG(if(!a) {error("Internal error in hash_qry(): use of uninitialized hash!!");})
    STAT(h_access__3++;)
    i32 idx=key%L;
    i32 k=a[idx].key;
    if(k==key || k==0) return &a[idx];
    i32 idx2=(key^0x7fffffff)%L;
    idx=idx2;
    i32 idxf=-1;
    do{
      STAT(h_collision__3++;)
      k=a[idx].key;
      if(k==key || k==0) return &a[idx];
      if(k<0 && idxf<0) idxf=idx;
      idx=(idx+1)%L;
    }while(idx!=idx2);
    if(idxf>=0) return &a[idxf];
    return 0;
  }
  i32 hash_data(i32 key){   //to be used for data reading, non-present entry returns 0    
    i32 idx=key%L;          //equivalent to {p=hash_qry(key); return p?(p->key>0?p->data:0):0;}
    DBG(if(!a) {error("Internal error in hash_data(): use of uninitialized hash!!");})
    STAT(h_access__3++;)
    i32 k=a[idx].key;
    if(k==key) return a[idx].data;
    if(k==0) return 0;
    i32 idx2=(key^0x7fffffff)%L;
    idx=idx2;
    do{
      STAT(h_collision__3++;)
      k=a[idx].key;
      if(k==key) return a[idx].data;
      if(k==0) return 0;
      idx=(idx+1)%L;
    }while(idx!=idx2);
    return 0;
  } 
  inline void hash_ins(entry__3* he, i32 key, i32 data) {occupied++; he->key=key; he->data=data;}
  inline void hash_del(entry__3* he) {occupied--; he->key=-1;}
  void rehash(F oversize){
    STAT(h_rehash__3++;)
    entry__3* b=a;
    int LL=L;
    init(oversize, occupied);
    for(int p=0; p<LL; p++){
      i32 k=b[p].key;
      if(k>0){
        entry__3* he=hash_qry(k);
        DBG(if(!he) error("Internal error in rehash(): out of prealloced space!!");)
        hash_ins(he, k, b[p].data);
      }
    }
    if(LL>0){
      free(b);  /* free it, provided it was not an empty hash */
      STAT(h_alloced_entries__3-=LL;)
    }
  }
};

STAT(i64 h_collision_build1__3; i64 h_access_build1__3; i64 h_size1__3; i64 h_hmem1__3;)
STAT(i64 h_collision_build2__3; i64 h_access_build2__3; i64 h_size2__3; i64 h_hmem2__3; i64 h_omem2__3;)

struct line__3{
  i32 map;
  i32 a,b;
  F Q;      /*  a=0  b=0  =>  Q=0 */
};

struct ctab__3{
  hash__3* c;
  int A,C,N;
  int B;      //number of unique bigrams
  i32* c_map;
  i32* alias;
  i32* epoch_guard;
  i32 epoch_ct;
  i32** x_walk;
  i32* x_walk_size;
  i32* c0, *c1;
  F* Rc0, *Rc1;
  F I;

#ifdef STATISTICS
  i64 htab_memory() { return sizeof(entry__3)*h_alloced_entries__3 + 
                             sizeof(hash__3)*A; }
  i64 other_memory() { 
    i64 acc=sizeof(i32)*A   // alias
           +sizeof(i32)*A   // epoch_guard
           +sizeof(i32)*C   // c_map
           +sizeof(i32*)*C  // x_walk
           +sizeof(i32)*C   // x_walk_size
           +sizeof(i32)*2*C // c0 and c1
           +sizeof(F)*2*C   // Rc0 and Rc1
           +sizeof(ctab__3);
    for(int k=0; k<C; k++)
      acc+=sizeof(i32)*x_walk_size[k];
    return acc;
  }
#endif

  i32 live(i32 y){
    i32 z=y; 
    if(alias[z]!=z){
      do{z=alias[z];}while(alias[z]!=z);
      do{i32 y1=alias[y]; alias[y]=z; y=y1;} while(alias[y]!=y);
    }
    return z;
  }   
  
  ~ctab__3(){
    for(int k=0; k<A; k++) c[k].done();
    free(c); free(c_map);
    for(int k=0; k<C; k++) dispose(x_walk[k]); 
    free(x_walk); free(x_walk_size);
    free(epoch_guard); free(alias);
    free(c0); free(c1); free(Rc0); free(Rc1);
    STAT(if(h_alloced_entries__3!=0) error("ctab__3(): Internal error: memory leak of %Ld bytes\n", h_alloced_entries__3*sizeof(entry__3));)
  }

  void build_tab(i32* w, i32 step, i32 span, F oversize, int* ysize){
    for(int k=0; k<N; k++){
      i32 y=w[step*k+span], x=w[step*k];
      if(c[y].L<0) c[y].init(oversize, ysize[y]);
      entry__3* p=c[y].hash_qry(x);
      DBG(if(!p) error("build_tab: Internal error, out of space, y=%d,x=%d\n", y,x);)
      if (p->key<=0) c[y].hash_ins(p,x,i32(0));
      p->data++;
    }
  }
  
  void load(i32* w, i32 NN, i32 step, i32 span, i32 Cmax, i32 T)
  {
    int max_w=0, M=0;
    for(int k=0; k<NN-span; k+=step, M++){
      if(max_w<w[k+span]) max_w=w[k+span];
      if(max_w<w[k]) max_w=w[k];
    }
    A=max_w+1; //we need space 1:max_w, so we allocate 0:max_w=[$A] where [0] will never be used
    N=M;
    alloc(c,A); for(int a=0; a<A; a++) c[a].init();
    int cw0[A], cw1[A];
    fill(cw0,i32(0),A); for(int k=0; k<NN-span; k+=step) cw0[w[k+span]]++;
    fill(cw1,i32(0),A); for(int k=0; k<NN-span; k+=step) cw1[w[k]]++;

    STAT(h_collision__3=0; h_access__3=0;)
    build_tab(w, step, span, 0.0, cw0);  //first build of ctab is only used to compute space demands (counts how many bigrams are there)
    STAT(h_collision_build1__3=h_collision__3; h_access_build1__3=h_access__3; 
         h_size1__3=h_alloced_entries__3; h_hmem1__3=htab_memory();)
    {
      int ct[A]; for(int a=0; a<A; a++) ct[a]=c[a].occupied;
      for(int a=0; a<A; a++) c[a].done(); //delete the table now
      STAT(if(h_alloced_entries__3!=0) error("ctab__3(): Internal error: memory leak of %Ld bytes\n", h_alloced_entries__3*sizeof(entry__3));
           h_collision__3=0; h_access__3=0;)
      build_tab(w,step,span,oversize,ct);         //and build the right-sized one
      B=0; for(int a=1; a<A; a++) B+=c[a].occupied;
      STAT(h_collision_build2__3=h_collision__3; h_access_build2__3=h_access__3; h_size2__3=h_alloced_entries__3;)
    }
    STAT(h_collision__3=0; h_access__3=0; h_rehash__3=0;)

    alloc(epoch_guard, A); epoch_ct=0;
    alloc(alias, A); for(int a=0; a<A; a++) alias[a]=a;

    C=0; int AA=A;
    for(int a=1; a<A; a++) if(cw0[a]+cw1[a]>=2*T) {
      C++; // AA=a+1;
  //    if(C==Cmax) break;
    }
AA=A; /// !!!! pozor tady je to blbe kdyby se pouzilo Cmax !!!!
    int inv_c_map[AA]; fill(inv_c_map, -1, AA);
    
    alloc(c_map,C);    
    for(int a=1,i=0; a<AA; a++) if(cw0[a]+cw1[a]>=2*T) {c_map[i]=a; inv_c_map[a]=i; i++;}

    alloc(x_walk,C); alloc(x_walk_size, C);
    {
      int ct[C]; fill(ct,0,C);
      for(int y=1; y<AA; y++) for(int p=0; p<c[y].L; p++){  //we can get here only if c[y].L>0 which is iff c[y]-th hash table is non-empty
        int x=c[y].a[p].key;
        if(x>0 && inv_c_map[x]>=0) ct[inv_c_map[x]]++;
      }
      for(int k=0; k<C; k++){
        alloc(x_walk[k],ct[k]);
        x_walk_size[k]=ct[k];
      }
    }
//printf("NN=%d, M=%d N=%d A=%d\n", NN, M, N, A);
    int last[C]; fill(last,0,C);
    I=0;
    for(int y=1; y<A; y++) for(int p=0; p<c[y].L; p++) {  //we can get here only if c[y].L>0 which is iff c[y]-th hash table is non-empty
      entry__3* e=&(c[y].a[p]);
      i32 x=e->key;
      if(x>0){
        i32 cyx=e->data;
        I-=cyx*ulog2( F(cw0[y])*cw1[x] / (F(N)*cyx) );          
        int xx=inv_c_map[x];
        if(xx>=0) x_walk[xx][last[xx]++]=y;
      }
    }
    I/=N;
    alloc(c0,C); alloc(c1,C); alloc(Rc0,C); alloc(Rc1,C);    
    for(int k=0; k<C; k++){
      c0[k]=cw0[c_map[k]]; Rc0[k]=R(c0[k]);
      c1[k]=cw1[c_map[k]]; Rc1[k]=R(c1[k]);
    }
    STAT(h_hmem2__3=htab_memory(); h_omem2__3=other_memory();)
  }

  ctab__3(i32* w, i32 NN, i32 step, i32 span, i32 Cmax, i32 T) { load(w,NN,step,span,Cmax,T); }

  void merge_columns(i32 ll, i32 rr){
    if(epoch_ct==0) fill(epoch_guard,i32(0),A);
    epoch_ct++;
    DBG(if(ll>=rr) error("merge_columns(): %d<%d violated\n", ll, rr);)
    int l=c_map[ll], r=c_map[rr];
    int bfr[x_walk_size[rr]], bp=0;
    i32* x_walk_rr=x_walk[rr];
    for(int yp=0, yp_to=x_walk_size[rr]; yp<yp_to; yp++){
      i32 y=x_walk_rr[yp];
      y=live(y);
      if(epoch_guard[y]==epoch_ct) continue;
      epoch_guard[y]=epoch_ct;           /*c[y].L>0*/  DBG(if(c[y].L<=0) error("Internal error in merge_columns(): c[%d].L=%d<=0", y, c[y].L);)
      entry__3* hr=c[y].hash_qry(r);    /*always finds*/ DBG(if(!hr || hr->key<=0) error("Internal error in merge_columns(): 'always finds' violated y=%d (was %d) x=%d key=%d\n",y, x_walk[rr][yp], r, hr? hr->key:-2);)
      i32 d=hr->data;
      c[y].hash_del(hr);
      entry__3* hl=c[y].hash_qry(l); /*always non-null*/ DBG(if(!hl) error("Internal error in merge_columns(): `always non-null' violated");)
      if(hl->key<=0) {c[y].hash_ins(hl,l,d); bfr[bp++]=y;}
      else{
        hl->data+=d;
        if(c[y].occupied*rehash_trigger < c[y].L) c[y].rehash(oversize);
      }
    }
    dispose(x_walk[rr]); x_walk_size[rr]=0;
    if(bp>0){
      i32* new_x_walk; alloc(new_x_walk, x_walk_size[ll]+bp);
      copy(new_x_walk, x_walk[ll], x_walk_size[ll]);
      copy(new_x_walk+x_walk_size[ll], bfr, bp);
      x_walk_size[ll]+=bp;
      free(x_walk[ll]);
      x_walk[ll]=new_x_walk;
    }
  }
   
  void merge_rows(i32 ll, i32 rr){
    DBG(if(ll>=rr) error("Internal error in merge_rows(): %d<%d violated\n", ll, rr);)
    int l=c_map[ll], r=c_map[rr];
    if(c[l].L<0) c[l].move(c[r]);  /* calls the move-constructor 'c[l] <-- c[r]' */
    else if(c[r].L>0){
      hash__3* lh=&(c[l]), *rh=&(c[r]);
      int intersection=0;  /* compute intersection size */
      if(lh->occupied<rh->occupied) swap(lh,rh);    
      for(int k=0, rL=rh->L; k<rL; k++) {
        i32 rkey=rh->a[k].key;
        if(rkey>0){
          entry__3 *er=lh->hash_qry(rkey);
          if(er && er->key>0) intersection++;
        }
      }
      int s=0, S=lh->occupied+rh->occupied;
      entry__3 bfr[S];        /* save values */
      for(int k=0, lL=lh->L; k<lL; k++) if(lh->a[k].key>0) bfr[s++]=lh->a[k];
      for(int k=0, rL=rh->L; k<rL; k++) if(rh->a[k].key>0) bfr[s++]=rh->a[k];
      lh->done(); rh->done();
      lh=&(c[l]);
      lh->init(oversize, S-intersection);   /* create the merged row */ 
      for(s=0; s<S; s++){
        entry__3* e=lh->hash_qry(bfr[s].key);   DBG(if(!e) error("Internal error in merge_rows(): e!=null violated");)
        if (e->key<=0) lh->hash_ins(e,bfr[s].key,bfr[s].data);
        else e->data+=bfr[s].data;
      }
    }
  }
   
  void merge(i32 ll, i32 rr)
  {
    merge_columns(ll,rr);
    merge_rows(ll,rr);
    c0[ll]+=c0[rr]; c1[ll]+=c1[rr];
    Rc0[ll]=R(c0[ll]); Rc1[ll]=R(c1[ll]);
    alias[c_map[rr]]=c_map[ll];
    c_map[rr]=0;
  }

  inline i32 cd(i32 y, i32 x) {return c[y].hash_data(x);}
  
  inline ux yx2i(int y, int x) { return y+ux(x)*(x-1)/2; }  /* y<x */


  void apply_corrections(FF* loss, line__3* line_ab, int n_ab,   /* both non-zero */
                                   line__3* line_a,  int n_a,    /* b-part is zero */
                                   line__3* line_b,  int n_b)    /* a-part is zero */
  {
    for(int p=0; p<n_ab; p++){
      i32 lpa=line_ab[p].a, lpb=line_ab[p].b; F lpQ=line_ab[p].Q; i32 lpm=line_ab[p].map;
      for(int q=0; q<n_ab; q++){ 
        i32 lqm=line_ab[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]+=lpQ+line_ab[q].Q-Q(lpa+line_ab[q].a, lpb+line_ab[q].b);
      }
      for(int q=0; q<n_a; q++){ 
        i32 lqm=line_a[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]+=lpQ-Q(lpa+line_a[q].a, lpb);
      }
      for(int q=0; q<n_b; q++){ 
        i32 lqm=line_b[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]+=lpQ-Q(lpb+line_b[q].b, lpa);
      }
    }

    for(int p=0; p<n_a; p++){
      i32 lpa=line_a[p].a; i32 lpm=line_a[p].map;
      for(int q=0; q<n_ab; q++){ 
        i32 lqm=line_ab[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]+=line_ab[q].Q-Q(lpa+line_ab[q].a, line_ab[q].b);
      }
      for(int q=0; q<n_b; q++){ 
        i32 lqm=line_b[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]-=Q(line_b[q].b, lpa);
      }
    }

    for(int p=0; p<n_b; p++){
      i32 lpb=line_b[p].b; i32 lpm=line_b[p].map;
      for(int q=0; q<n_ab; q++){ 
        i32 lqm=line_ab[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]+=line_ab[q].Q-Q(line_ab[q].a, lpb+line_ab[q].b);
      }
      for(int q=0; q<n_a; q++){ 
        i32 lqm=line_a[q].map;
        if(lpm<lqm)
          loss[yx2i(lpm,lqm)]-=Q(line_a[q].a, lpb);
      }
    }    
  }

  void I_corrections(FF* loss, i32 ll, i32 rr)
  {
    i32 l=c_map[ll], r=c_map[rr];
    line__3 line_ab[C]; 
    line__3 line_a[C]; 
    line__3 line_b[C]; 
    
    int n_ab=0, n_a=0, n_b=0;              /*!!!!!!! tady budu potrebovat inv_cmap !!! */
    for(int p=0; p<C; p++) if(p!=ll && p!=rr && c_map[p]>0) {
      i32 a=cd(l,c_map[p]), b=cd(r,c_map[p]);    
      if(a!=0 && b!=0) {
        line_ab[n_ab].a=a; line_ab[n_ab].b=b; 
        line_ab[n_ab].Q=Q(a,b); line_ab[n_ab].map=p;
        n_ab++;
      }else if(a!=0){
        line_a[n_a].a=a; line_a[n_a].map=p;
        n_a++;
      }else if(b!=0){
        line_b[n_b].b=b; line_b[n_b].map=p;
        n_b++;
      }
    }
    apply_corrections(loss, line_ab, n_ab, line_a, n_a, line_b, n_b);
    
    n_ab=0, n_a=0, n_b=0;
    for(int p=0; p<C; p++) if(p!=ll && p!=rr && c_map[p]>0) {
      i32 a=cd(c_map[p], l), b=cd(c_map[p], r);
      if(a!=0 && b!=0) {
        line_ab[n_ab].a=a; line_ab[n_ab].b=b; 
        line_ab[n_ab].Q=Q(a,b); line_ab[n_ab].map=p;
        n_ab++;
      }else if(a!=0){
        line_a[n_a].a=a; line_a[n_a].map=p;
        n_a++;
      }else if(b!=0){
        line_b[n_b].b=b; line_b[n_b].map=p;
        n_b++;
      }
    }
    apply_corrections(loss, line_ab, n_ab, line_a, n_a, line_b, n_b);
  }

  F I_loss(i32 l, i32 r)
  {
    if(epoch_ct==0) fill(epoch_guard,i32(0),A);
    epoch_ct++;
    int ml=c_map[l], mr=c_map[r];
    i32 cll=cd(ml,ml), clr=cd(ml,mr), crl=cd(mr,ml), crr=cd(mr,mr);
    F base = R(c0[l]+c0[r]) - Rc0[l] - Rc0[r] + R(c1[l]+c1[r]) - Rc1[l] - Rc1[r]
           + R(cll+clr) + R(crl+crr) + R(cll+crl) + R(clr+crr)
           - R(cll)-R(clr)-R(crl)-R(crr)-R(cll+clr+crl+crr);
    F acc=0;
    if(c[ml].occupied < c[mr].occupied) {swap(ml,mr);swap(l,r);}
    for(int p=c[mr].L-1; p>=0; p--){
      i32 x=c[mr].a[p].key;
      if(x>0){
        i32 clx=cd(ml,x);
        if(clx>0){              /* update is non-zero in the intersection only */
          F crx=c[mr].a[p].data;
          F alfa=clx/F(clx+crx), beta=1-alfa;   /* beta=crx/(clx+crx) */
          acc+=clx*ulog2(alfa) + crx*ulog2(beta);  
        }
      }
    }    

    if(x_walk_size[l]<x_walk_size[r]) {swap(l,r);swap(ml,mr);}
    int q=0;
    for(int p=0; p<x_walk_size[r]; p++){
      int y=live(x_walk[r][p]);
      y=live(y);
      if(epoch_guard[y]==epoch_ct) continue;
      epoch_guard[y]=epoch_ct;
      x_walk[r][q++]=y;      
      int cyl=cd(y,ml);
      if(cyl>0){
        int cyr=cd(y,mr);
        F alfa=cyl/F(cyl+cyr), beta=1-alfa;
        acc+=cyl*ulog2(alfa) + cyr*ulog2(beta); 
      }
    }
    x_walk_size[r]=q;
    return base+acc;
  }
};

void mmi_clustering_3(i32* w, i32 N, i32 step, i32 span, i32 Cmax, i32 T, mmi_result &ret, FILE* log)
{
  STAT(int start_time=time(0);)
  VERB(fprintf(stdmsg,"Building table for data of length %d with threshold %d\n", N, T);)
  ctab__3 c(w,N,step,span,Cmax,T);
  VERB(fprintf(stdmsg," ... done A=%d B=%d C=%d\n", c.A, c.B, c.C);)
  tree t000; t000.l=0; t000.r=0; t000.level=0;
  alloc(ret.forest, c.A+c.C-1); fill(ret.forest, t000, c.A);
  alloc(ret.loss, c.C-1);
  ret.I=c.I; ret.A=c.A; ret.B=c.B; ret.C=c.C;
#ifdef PROGRESS_LOG  
  F totI=ret.I;
  if(log) {
    fprintf(log,"#%5s %30s %30s %5s\n", "index", "MI_loss", "MI", "level");
    fprintf(log,"%6d %30.28f %30.28f %5d\n", c.A-1, 0.0, totI, 0);
  }
#endif
  int out_map[c.C]; copy(out_map, c.c_map, c.C);
  FF loss[ux(c.C-1)*c.C/2];  /* to be indexed by yx2i */
  for(int x=1; x<c.C; x++) for(int y=0; y<x; y++) loss[c.yx2i(y,x)]=c.I_loss(y,x);
  VERB(fprintf(stdmsg,"initial loss table computed, starting iterations...\n"));
  for(int n=c.A; n<c.A+c.C-1; n++){
    int l=0,r=1;
    FF min_loss=F_max(FF(1));
    for(int x=1; x<c.C; x++) if(c.c_map[x]>0) for(int y=0; y<x; y++) if(c.c_map[y]>0 && min_loss>loss[c.yx2i(y,x)]){
      min_loss=loss[c.yx2i(y,x)];
      l=y; r=x;
    }
    VERB(fprintf(stdmsg,"step %d of %d: merging %d %d  --> %d            \r", n-c.A+1, c.C-1, out_map[l], out_map[r], n));
    c.I_corrections(loss,l,r);
    c.merge(l,r);
    ret.forest[n].l=out_map[l]; ret.forest[n].r=out_map[r]; ret.loss[n-c.A]=min_loss/c.N;
#ifdef PROGRESS_LOG
    if(log){
      ret.forest[n].level=1+max(ret.forest[out_map[l]].level, ret.forest[out_map[r]].level);
      totI-=min_loss/c.N;
      fprintf(log,"%6d %30.28f %30.28f %5d\n", n, min_loss/c.N, totI, ret.forest[n].level);
      fflush(log);
    }
#endif
    for(int y=0; y<c.C; y++) if(y!=l && c.c_map[y]>0){
      int p=y, q=l;
      if (p>q) swap(p,q);
      loss[c.yx2i(p,q)]=c.I_loss(p,q);
    }
    out_map[l]=n;
  }
  VERB(fprintf(stdmsg,"\n");)
#ifdef STATISTICS
  VERB(fprintf(stdmsg,
       "mmi_clustering_3(): Finished in %d seconds. \n"
       " probe hash table's insert count = %Ld, collisions per insert = %4.2f\n"
       "      allocated entry count = %Ld,  size=%4.2f MB\n"
       " main hash table's insert count = %Ld, collisions per insert = %4.2f\n"
       "      allocated entry count = %Ld,  size=%4.2f + %4.2f MB\n"
       "      access count = %Ld, collisions per access = %4.2f\n"
       "      rehash count = %Ld\n",
       int(time(0))-start_time,
       h_access_build1__3, D(h_collision_build1__3)/h_access_build1__3,
       h_size1__3, D(h_hmem1__3)/1024/1024,
       h_access_build2__3, D(h_collision_build2__3)/h_access_build2__3,
       h_size2__3, D(h_hmem2__3)/1024/1024, D(h_omem2__3)/1024/1024,
       h_access__3, D(h_collision__3)/h_access__3, h_rehash__3);)
#endif
}

void mmi_clustering_3(i32* w0, i32* w1, int N, int T, mmi_result &ret)  //for compatibility with `check'
{
  i32 w[2*N];
  for(int k=0; k<N; k++) {w[2*k]=w0[k]; w[2*k+1]=w1[k];}
  mmi_clustering_3(w,2*N,2,1,1<<30,T,ret,0);
}

