// >>>>> mmi_tree.cc

#define ERROR_MSG_PREFIX "mmi_tree: "
#include "mmi_clustering_4.cc"

char** abc;

char* bp="sbp--sliding%s";
int W=3000;
char* outname="out";
char* blocks=0;
int no2phase_sw=0;
int* eqmap;

int main(int argc, char**argv)
{
  int help=0, err=0;
  argv++; argc--;
  while(*argv && argv[0][0]=='-'){
    char* t;
    if(strcmp(*argv,"--help")==0 || strcmp(*argv,"-h")==0) help=1;
    else if(strcmp(*argv,"--bp")==0){
      argv++; argc--;
      int l=0; bp=0;
      while(*argv && strcmp(*argv, "--")!=0){
        int m=strlen(*argv);
        resize(bp,l+1+m);
        copy(bp+l,*argv,m);
        l+=m+1;
        bp[l-1]=''; bp[l]=0;
        argv++; argc--;
      }
      if(l==0) error("preprocessor name expected");
      if(!*argv) error("unterminated preprocessor call (`--' expected)");
      bp[l-1]=0;
    }else if((t=is_prefix("--window=", *argv))){
      char* endptr;
      W=strtol(t,&endptr,10);
      if(*endptr!=0) error("number expected after `--window='");
      if(W<=10) error("--window has to be greater than 10");
      if(W==INT_MAX) error("--window too large");
    }else if((t=is_prefix("--blocks=", *argv))) blocks=t;
    else if((t=is_prefix("--out=", *argv)))     outname=t;
    else if(strcmp("--no2phase=", *argv)==0) no2phase_sw=1;
    else error("unknown option %s", *argv);
    argv++; argc--;
  }
  check_ps(bp);
  if(!*argv) err=1;
  if(help || err){
    printf(
"MMI-tree builder                                       mmi_clustering v"VERSION"\n"
"Usage: mmi_tree OPTIONS ( binfile_i )+\n"
"Input is read from binfiles, the output will be written into\n"
"  outname.htree   human readable tree\n"
"  outname.tree    machine representation of the tree\n"
"OPTIONS alter behavior in the following way (defaults are in brackets):\n"
"  --bp P A1 ... An -- specifies preprocessor call. Preprocessor P is called\n"
"         with arguments A1 to An. Any occurrence of %%s in Ai (up to 20 of \n"
"         them are valid) will be replaced by an input file name. Preprocessor\n"
"         has to write each observation to stdout in a binary form as a triple\n"
"         (leftword,rightword,count)          [--bp sbp --sliding %%s --]\n"
//"  --window=W  at most W-words can be considered at once [W=3000]\n"
"  --window=W  at most W-words will be clustered [W=3000]\n"
"  --out=O  output name prefix (can contain directory name which will\n"
"           be eventually created)                         [O=out]\n"
/*"  --blocks=F if specified, file F with equivalence classes serves as a\n"
"             seed. In the first phase, only words inside these classes can\n"
"             be merged to form internal nodes of the tree (and cannot be\n"
"             merged across the classes). Once all the classes has shrunk\n"
"             to singletons, second phases is allowed to merge any word with\n"
"             any other to finish the tree. Without this option only the second\n"
"             will be run.\n"
"  --no2phase if --blocks has been specified, don't run second phase.\n"*/
"Example:\n"
"mmi_tree --out=example --bp sbp --span=3 --weight %%s -- bindir/file.bin\n"
);
    return err;
  }

  char bindir[strlen(*argv)+2]; strcpy(bindir, *argv);
  char* bs=strrchr(bindir,'/'); if (!bs) strcpy(bindir,"."); else *bs=0;
  int A=dictionary_size(bindir);
  if(blocks){
    load_dictionary(abc, bindir);
    eqmap=load_eqmap(blocks, abc, 0, A);
    free(abc); //save space before starting the clustering
  }
  
  FILE* out_htree, *out_log=0;
#ifdef PROGRESS_LOG
  FILE* out_hist;
#endif
  {
    char fname[strlen(outname)+10];
    sprintf(fname,"%s.htree", outname);
    out_htree=mkdir_fopen(fname, "w", 0644);
    if(!out_htree) error("Cannot open output file %s (%m)", fname);
#ifdef PROGRESS_LOG
    sprintf(fname,"%s.log", outname);
    out_log=mkdir_fopen(fname, "w", 0644);
    if(!out_log) error("Cannot open output file %s (%m)", fname); 
    sprintf(fname,"%s.hist", outname);
    out_hist=mkdir_fopen(fname, "w", 0644);
    if(!out_hist) error("Cannot open output file %s (%m)", fname); 
#endif
  }
  stdmsg=stderr; mmi_result r;
  mmi_clustering__4(out_log, bp, argv, bindir, A, W, r);


  { 
    char fname[strlen(outname)+10];
    sprintf(fname,"%s.tree", outname);
    save_tree(fname,bindir,r);
  }
       
  if(blocks) 
    fprintf(out_htree,"Preconstrained by BLCK0=%s\n",blocks);

  load_dictionary(abc, bindir);      
  i64* hist; load_histogram(hist, bindir);
  draw_tree(out_htree,PRINT_ILOSS|PRINT_FREQS, r,abc, hist, eqmap?&eqmap:0, 0, eqmap?1:0);
  if(fclose(out_htree)<0) error("cannot write (%m)"); 
#ifdef PROGRESS_LOG
  print_tree(out_hist,r,abc);  
  if(fclose(out_log)<0) error("cannot write (%m)"); 
  if(fclose(out_hist)<0) error("cannot write (%m)");
#endif
  return 0;
}

