// >>>>> os.cc

/* ----- this file comes from `daves_core' package'
 */
  

#ifndef OS_CC
#define OS_CC

#include <osbase.cc>
#include <sys/vfs.h>


FILE *logfile=0;

#ifndef ERROR_MSG_PREFIX
# define ERROR_MSG_PREFIX "Error: "
#endif

static int vlprintf(char* msg, va_list va) UNUSED;
static int vlprintf(char* msg, va_list va)
{
  if(logfile) vfprintf(logfile, msg, va);
  return vprintf(msg, va);
}

static int lprintf(char* msg, ...) UNUSED;
static int lprintf(char* msg, ...)
{
  va_list va; 
  va_start(va, msg);
  int r=vlprintf(msg, va);
  va_end(va);
  return r;
}

static void error(char* msg, ...) UNUSED NORETURN;

static int get_mtime(char* path) UNUSED;
static int get_mtime(char* path)
{
  struct stat st;
  if(lstat(path,&st)<0) error("get_mtime(): cannot stat file %s (%s)", path, strerror(errno));
  return st.st_mtime;
}

static int file_size(char* path) UNUSED;
static int file_size(char* path)
{
  struct stat st;
  if(lstat(path,&st)>=0)
    if(S_ISREG(st.st_mode)) return st.st_size;
    else error("file_size(): file `%s' should be regular", path);
  return 0;
}

static u64 disk_free(char* path) UNUSED;
static u64 disk_free(char* path) 
{
  struct statfs st;
  if(statfs(path,&st)<0) error("disk_free(): cannot statfs `%s' (%m)", path);
  return st.f_bavail * u64(st.f_bsize);
}

/*
 Vrati maximalni mtime ktery se vyskytuje v adresari path (vcetne jeho
 sameho), nebo hodnotu parametru mtime, podle toho co je vetsi. Parametr
 path musi umet pojmout PATH_MAX znaku a bude menen.
*/
static int max_dir_mtime(char* path, int mtime) UNUSED;
static int max_dir_mtime(char* path, int mtime) 
{
  struct stat st;  
  if(lstat(path, &st)<0) return mtime;
  mtime=max(mtime, int(st.st_mtime));
  if(S_ISDIR(st.st_mode)){
    DIR* d=opendir(path); int PL=norm_path(path);
    if(path[PL-1]!='/') path[PL++]='/';
    dirent*  de;
    while( (de=readdir(d))!=0 ){
      char* dn=de->d_name;
      if(strcmp(".", dn)==0 || strcmp("..", dn)==0) continue;
      int DL=strlen(dn);
      if(PL+DL>=PATH_MAX) error("max_dir_mtime(): out of PATH buffer for %s + %s", path, dn);
      strcpy(path+PL, dn);
      mtime=max_dir_mtime(path, mtime);
    }
    closedir(d);
  }
  return mtime;
}


static void handle_child_err(int a, int b) UNUSED NORETURN;
static void handle_child_err(int a, int b)
{
  int e=errno;
  int tty=open("/dev/tty", O_RDWR);
  if(tty>=0){
    FILE* out=fdopen(tty, "w");
    if(out)
      fprintf(out,"Cannot spawn() child process (%s)(%d)(%d)!!! \n"
                  "This (un)expected error was not handled because the ill-designed\n"
                  "UNIX makes it extremely difficult and the author was extremely lazy.\n", strerror(e),a,b);
  }
  exit(121);
}

/*
  Runs `command' in a directory `in_directory' eventually printing a message
  upon its start if disp!=0. The command is run with all its filedescriptors
  closed, except those in `filedes' array which are remapped to
  0..length(filedes) for commands' sake. Array `filedes' must be ended by
  -1. The `command' string is first processed in a printf fashion (using
  va_list arguments), to obtain string which will then be chopped into the
  program and its arguments at places denoted by '' character. The
  executable of the program is searched using $PATH environment variable,
  unless it contains '/' character (in this case the executable name is
  treated as an absolute path).
  
  Example:  
    vrun("..", 0, (int[]){0,1,1,-1}, "configure%s--disable-nls", arg);
   
  Returns pid of the spawned process or -1 on error (and sets errno), or -2 on 
  hard error.
*/
static int vrun(char* in_directory, int disp, int* filedes, char* command, va_list va) UNUSED;
static int vrun(char* in_directory, int disp, int* filedes, char* command, va_list va)
{
  int base_pid=getpid(); DIR* d;  //nejdriv se zjisti co vse je otevreno aby to mohl child pozavirat
  char procdir[64];
  sprintf(procdir, "/proc/%d/fd/", base_pid);   //tohle je taky blby protoze to nejde cist kdyz je proces mod 711, own root a ten kdo ho spusti non-root !!!
  if((d=opendir(procdir))==NULL) return -1;
  int OF=0;
  while(readdir(d)) OF++;
  int open_files[OF+1];  //+1 proto ze int x[0] na starsich gcc myslim delalo problemy
  rewinddir(d);
  dirent* de; OF=0;
  while((de=readdir(d))) if(isdigit(de->d_name[0])) open_files[OF++]=atoi(de->d_name);
  closedir(d);

  char* cmd; int L=strlen(command)*2;
  for(;;){  //taky trochu divny, ale jak to delat jinak?
    alloc(cmd,L);  
    if(vsnprintf(cmd, L, command, va)<L-1) break;
    free(cmd); L*=2;  //krome toho man page tvrdi ze kdyz se to tam nevejde tak vrati -1
  }                   //jenze tady se to chova tak ze to vrati delku kterou mam alokovat !?
  int sep=0;
  for(int i=0; cmd[i]; i++) if(cmd[i]=='') sep++;
  char *args[sep+2]; args[0]=cmd; args[sep+1]=0;
  for(int i=0, k=1; cmd[i]; i++) if(cmd[i]==''){
    args[k++]=cmd+i+1;
    cmd[i]=0;
  }
  if(disp){
    for(int i=0; args[i]; i++) lprintf("%s ", args[i]);
    lprintf("\n");
  }
  fflush(stdout);

    
  char cwd[PATH_MAX+1];
  getcwd(cwd,PATH_MAX);  //tady bych mel spis testovat jestli tam muzu chdir()nout a ne to primo zkouset
  if(chdir(in_directory)<0) {free(cmd); return -1;}  //ale kdo by se s tim psal kdyz by to problem stejne uplne nevyresilo (viz. nize)
  char exe[PATH_MAX+1];  //krome toho se mi to zde hodi, protoze kdyz hledam pres path, musim dodrzovat i '.' abych spravne hledal i relativne ulozene programy
  if(strchr(cmd, '/'))   //search cmd in PATH only if it does not contain '/'
    strcpy(exe,cmd);
  else{
    char* path=getenv("PATH");
    if(!path) path=":/bin:/usr/bin";
    char* end=path+strlen(path);
    while(path<end){
      char* p=strchr(path, ':');
      int L = (!p) ? strlen(path) : p-path;
      if(L>0){
        copy(exe, path, L);
        exe[L]='/';
        strcpy(exe+L+1, cmd);
        if(can_exec(exe)) break;
      }
      path+=L+1;
    }
  }
  if(chdir(cwd)<0) {free(cmd); return -2;}  //-2 == lost cwd -- maly nemily dusledek predchozi lenosti
  if(!can_exec(exe)) {free(cmd); errno=ENOENT; return -1;}

  int pid=fork();                       // --- FORK -----------------
  if(pid<0) {free(cmd); return -1;}
  if(pid==0){  //--child--
    for(int f=1; f<NSIG; f++) signal(f, SIG_DFL);
    if(chdir(in_directory)<0) handle_child_err(0,0);
    int N=0; while(filedes[N]>=0) N++;
    int copy_area[N+1],  //copy_area umozni descriptorum prezit eru zavirani
        force_area[N+1]; //a force_area je tam proto aby copy_area[x]>=N x
    for(int i=0; i<N; i++) if((force_area[i]=dup(filedes[i]))<0) handle_child_err(1,i);
    for(int i=0; i<N; i++) if((copy_area[i]=dup(filedes[i]))<0) handle_child_err(2,i);
    for(int i=0; i<N; i++) if(close(force_area[i])<0) handle_child_err(3,i); //Neprehledna politicka situace?
    for(int i=0; i<OF;i++) close(open_files[i]);                             //  ---> pozavirat vsechny cizince!    (ignore errors for close-on-exec)
    for(int i=0; i<N; i++) if(dup2(copy_area[i],i)<0) handle_child_err(4,i); //z tech co zustali vyrobime spolehlive soudruhy
    for(int i=0; i<N; i++) if(close(copy_area[i])<0) handle_child_err(5,i);  //ale az splni svou roli tak se jich zbavime taky
    execv(exe, args);                                                        //tim bude revoluce descriptoru zavrsena
    handle_child_err(6,0);//tohle je v UNIXu fakt hnusny. Vlastne nejde poznat (bez spoluprace s
  }                       //`cmd') ze se program uspesne spustil. Muzu nanejvys cekat 
  else { // --parent--    //a kdyz to do nejaky doby nerachne tak asi bezi. Jenze to by
    free(cmd);            //zase brzdilo parenta. Krome toho neni vubec snadny z childa
    return pid;           //predat chybovy kod do parenta. Proto se snazim co nejvic chyb
  }                       //odhalit pred fork()em, a kdyz pak dojde k chybe tak child ex()ne.
}

#define CHILD_KILLED (1<<30)

/*
Similar to vrun(), but waits for the child to terminate.
Returns -1 or -2 as an error, or exit status which is either a number with CHILD_KILLED bit
cleared, or set. In the first case, the lower bits represent exit()-code, in the later case
these bits represent the number of terminating signal.
*/
static int vcall(char* in_directory, int disp, int* filedes, char* command, va_list va) UNUSED;
static int vcall(char* in_directory, int disp, int* filedes, char* command, va_list va)
{
  int pid=vrun(in_directory, disp, filedes, command, va);
  if(pid<0) return pid;
  int status;
  for(;;){
    int r=waitpid(pid, &status, 0);  //musi to byt v cyklu protoze podle man page
    if(r==pid) break;                //to chapu tak ze kdyz mi prijde signal tak se to 
    if(r<0) return -2;               //bezduvodne probudi
  }
  if(WIFEXITED(status)) return WEXITSTATUS(status);
  else return WTERMSIG(status)|CHILD_KILLED;
}


/*
The same as vcall() but prints appropriate error messages and exits whenever
anything goes wrong.
*/
void vecall(char* in_directory, int disp, int* filedes, char* command, va_list va) UNUSED;
void vecall(char* in_directory, int disp, int* filedes, char* command, va_list va)
{
  if(disp) lprintf("RUN(%s): ", in_directory);
  int status=vcall(in_directory, disp, filedes, command, va);
  if(status==-1) error("vecall(): could not call `%s' (%s)", command, strerror(errno));
  if(status==-2) error("vecall(): internal error in waitpid() while calling `%s' (%s)", command, strerror(errno));
  if(status & CHILD_KILLED) error("vecall(): child `%s' killed by signal %d", command, status^CHILD_KILLED);
  if(status!=0) error("vecall(): child `%s' exited with error code %d", command, status);
}

/*
`...' versions of the preceding functions
*/
static int run(char* in_directory, int disp, int* filedes, char* command, ...) UNUSED;
static int run(char* in_directory, int disp, int* filedes, char* command, ...) 
{
  va_list va; va_start(va, command);
  int r=vrun(in_directory, disp, filedes, command, va);
  va_end(va);
  return r;
}
static int call(char* in_directory, int disp, int* filedes, char* command, ...) UNUSED;
static int call(char* in_directory, int disp, int* filedes, char* command, ...)
{
  va_list va; va_start(va, command);
  int r=vcall(in_directory, disp, filedes, command, va);
  va_end(va);
  return r;
}
static void ecall(char* in_directory, int disp, int* filedes, char* command, ...) UNUSED;
static void ecall(char* in_directory, int disp, int* filedes, char* command, ...)
{
  va_list va; va_start(va, command);
  vecall(in_directory, disp, filedes, command, va);
  va_end(va);
}
  
static void mkdirs(char* path, int mode) UNUSED;
static void mkdirs(char* path, int mode)
{
  char p[strlen(path)+1]; strcpy(p,path);  //aby to fungovalo i pro const char*  
  if(rec_mkdir(p, mode)<0) error("mkdirs(): cannot make directory `%s' (%s)", path, strerror(errno));
}

static int vdelr(int disp, char* dir, va_list va) UNUSED;
static int vdelr(int disp, char* dir, va_list va)
{
  char d[PATH_MAX+1];
  vsnprintf(d, PATH_MAX, dir, va);
  if(disp) lprintf(" DELETING directory %s\n", d);
  return call(".", 0, (int[]){-1}, "rm-rf%s", d);
}

static int delr(int disp, char* dir, ...) UNUSED;
static int delr(int disp, char* dir, ...)
{
  va_list va; 
  va_start(va, dir);
  int r=vdelr(disp, dir, va);
  va_end(va);
  return r;
}

static void bailout() UNUSED NORETURN;

static void verror(char* msg, va_list va) UNUSED NORETURN;
static void verror(char* msg, va_list va)
{
  fprintf(stderr,ERROR_MSG_PREFIX);
  vfprintf(stderr,msg, va);
  fprintf(stderr,"\n");
  bailout();
}

static void error(char* msg, ...) UNUSED NORETURN;
static void error(char* msg, ...)
{
  va_list va; 
  va_start(va, msg);
  verror(msg, va);
  va_end(va);
} 

#endif

