// >>>>> osbase.cc

/* ----- this file comes from `daves_core' package'
 */

#ifndef OSBASE_CC
#define OSBASE_CC

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>                      
#include <dirent.h>
#include <stdarg.h>
#include <signal.h>
#include <ctype.h>
#include <inc.h>


static int xread(int fd, char* buffer, int L) UNUSED;
static int xread(int fd, char* buffer, int L)
{
  int p=0;
  while(L>0){
    int l=read(fd, buffer+p, L);
    if(l<0) return l;
    if(l==0) return p;
    p+=l; L-=l;
  }
  return p;
}

template <class T> static inline int tread(int fd, T* buffer, int L)
{
  int r=xread(fd, (char*)buffer, sizeof(T)*L);
  if (r>=0) return r/sizeof(T);
  else      return r;
}
  

static int xwrite(int fd, char* buffer, int L) UNUSED;
static int xwrite(int fd, char* buffer, int L)
{
  int p=0;
  while(L>0){
    int l;
    do l=write(fd, buffer+p, L);
    while(l==-1 && errno==EINTR);
    if(l<0) return l;
    if(l==0) return p;
    p+=l; L-=l;
  }
  return p;
}

template <class T> static inline int twrite(int fd, T* buffer, int L)
{
  int r=xwrite(fd, (char*)buffer, sizeof(T)*L);
  if (r>=0) return r/sizeof(T);
  else      return r;
}

static int file_exists(char* path) UNUSED;
static int file_exists(char* path)
{
  struct stat st;
  return lstat(path,&st)>=0 && S_ISREG(st.st_mode);
}

static int directory_exists(char* path) UNUSED;
static int directory_exists(char* path)
{
  struct stat st;
  return lstat(path,&st)>=0 && S_ISDIR(st.st_mode);
}

/* normalizuje cestu (in situ) a vrati jeji delku Nejprve odstrani nasobne
 * vyskyty znaku '/', takze adresare cesty budou oddeleny prave jednim
 * znakem '/'. Navic se vyhodi leading `./'s. Potom nasleduje druhy pruchod,
 * pri kterem se vyhazuje `/./', a interpretuje `X/../' (kdyz nejde jit
 * zpet, zustane `..' ve vystupu, jako v pripade `/..' ktery prejde na
 * `/..'). Potom se odstrani pripadny trailing slash. Nakonec se nahradi ""
 * by "." ("" mohl vyjit behem vyhodnocovani `x/..').
 *
 * NOTE: This norm_path NEVER outputs relative path as `./something' (it
 * would use `something' instead).
 */
static int norm_path(char* path) UNUSED;
static int norm_path(char* path) 
{  
  if(path[0]==0) return 0;    //ignore non-path
  int i=0, j=0;
  while(path[j]){             //remove slash-runs
    char z=path[j++];
    path[i++]=z;
    if(z=='/')
      while(path[j]=='/') j++;
  }
  path[i]=0;    //here i>0 and it is a length of `path' string

  int start[2+i/2], p=-1, guard=1; //start[p] points to the beginning of p-th component of the path
  start[++p]=0; i=0; j=0;          //last valid component is the p-th one (note that there is zero-length component in case of root directory beginning the path)
  while(path[j]=='.' && path[j+1]=='/') j+=2; //remove leading './'s
  while(path[j]){   //remove '*/./*' and '*/X/../*' (but keep '/../..')
    char z=path[j++];
    path[i++]=z;
    if(z=='/'){
      start[++p]=i;
      while(path[j]=='.'){
        if(path[j+1]=='/') {j+=2; continue;}
        else if(path[j+1]==0) goto break_all;
        else if(path[j+1]=='.' && (path[j+2]=='/' || path[j+2]==0)){
          if(i>guard){
            i=start[--p];
            if(path[j+2]=='/') {j+=3; continue;}
            else goto break_all;
          }else{
            guard=i+3; 
            break;
          }
        }else break;
      }
    }
  }
break_all:
  if(i>1 && path[i-1]=='/') i--; //remove possible trailing slash (except if it was signle '/' -- the root directory)
  if(i==0) {path[0]='.'; i=1;}   //empty string --> current directory
  path[i]=0;
  return i;
}

/* Similar to norm_path() but always returns relative paths as `./something'.
 */
static int norm_path2(char* in, char* &out) UNUSED;
static int norm_path2(char* in, char* &out) 
{
  int L=norm_path(in);
  if(in[0]=='/') {out=strdup(in); return L;}
  alloc(out, L+3); in[0]='.'; in[1]='/'; strcpy(in+2,out);
  return L+2;
}

/* Vytvori absolutni cestu. IN: path (nezmeni ho), OUT: abs (alokuje ho)
 * Predpoklada ze 
 *  ~ -> $HOME,  ~root -> /root,  ~name -> /home/name
 * coz neni uplne dobre protoze /etc/password umoznuje mit home kdekoliv,
 * avsak prohledavat ho a hleda usera mi prijde prilis namahavy, takze zatim to je takhle
 */
static int abs_path(char* path, char* &abs) UNUSED;
static int abs_path(char* path, char* &abs)
{
  if(path[0]=='~'){
    char* bfr;
    if(path[1]=='/' || path[1]==0){
      bfr=getenv("HOME");
      if(!bfr) bfr="Error(abs_path)._HOME_must_be_set.";
      path++;
    }else if(is_prefix("~root", path) && (path[5]==0 || path[5]=='/')){
      bfr="/root"; path+=5;
    }else{
      bfr="/home/"; path++;
    }
    int N=strlen(bfr)+strlen(path)+1;
    alloc(abs,N);
    sprintf(abs,"%s%s",bfr,path);
  }else if(path[0]!='/'){
    char bfr[PATH_MAX+1];
    getcwd(bfr, PATH_MAX);
    int N=strlen(bfr)+strlen(path)+4;
    alloc(abs,N);
    sprintf(abs,"%s/%s",bfr,path);
  }
  else{
    alloc(abs,strlen(path)+1);
    strcpy(abs,path);
  }
  return norm_path(abs);
}

/* Returns >0 iff the file exits and I own it (traces symlinks)
 */
static int i_own(char* file) UNUSED;
static int i_own(char* file)
{
  struct stat st;
  if(stat(file, &st)<0) return 0;
  return st.st_uid==getuid();
}

/* Returns >0 iff the file exists and I can execute it (traces symlinks)
 */
static int can_exec(char* file) UNUSED;
static int can_exec(char* file)
{
  struct stat st;  int r=0;
  if(stat(file, &st)<0) return 0;
  if(!S_ISREG(st.st_mode)) return 0;
  if(getuid()==st.st_uid) r|=st.st_mode&S_IXUSR;
  if(getgid()==st.st_gid) r|=st.st_mode&S_IXGRP;
  r|=st.st_mode&S_IXOTH;
  return r;
}

  
static int mkdir0(char* name, int mode)
{
  int ret=mkdir(name, mode);
  if(ret>=0) return 0;
  else if(errno==EEXIST) {errno=0; return 0;}
  else return -1;
}

static int rec_mkdir(char* name, int mode)
{
  if(mkdir0(name, mode)==0) return 0;
  char* ptr=strrchr(name, '/');
  if(!ptr) return -1;
  *ptr=0;
  if(rec_mkdir(name, mode)<0) return -1;
  *ptr='/';
  return mkdir0(name, mode);
}


#endif

