// >>>>> sbp.cc

#define ERROR_MSG_PREFIX "sbp: "
#include "base.cc"
#define BLEN 8192

int sliding_sw=0,weight_sw=0,dump_sw=0,type_sw=0,S=1;

inline void emit(int* l, int *r)
{
  int ec=0;
  ec|=put32(stdout, *l);
  ec|=put32(stdout, *r);
  if(!dump_sw) ec|=put32(stdout, weight_sw ? S-abs((r-1-l)/2) : 1 );
  if(ec<0) error("cannot write (%m)");
}

inline int read_block(int f, int* bfr, int L)
{
  if(sliding_sw){
    bfr++;
    int l=tread(f,bfr, L);
    for(int p=l-1; p>=0; p--) bfr[2*p+1] = bfr[2*p] = bfr[p];
    if(l<0) error("cannot read (%m)");
    return l;
  }else{
    int l=tread(f,bfr, 2*L);
    if(l<0) error("cannot read (%m)");
    if(l&1) error("odd size read -- invalid input file");
    return l/2;
  }
} 

int main(int argc, char** argv)
{
  int help=0;  
  argv++; argc--;
  while(*argv && argv[0][0]=='-'){
    char* Sstr;
    if(strcmp(*argv,"--help")==0 || strcmp(*argv,"-h")==0) help=1;
    else if(strcmp(*argv,"--sliding")==0) sliding_sw=1;
    else if(strcmp(*argv,"--weight")==0) weight_sw=1;
    else if(strcmp(*argv,"--dump")==0) dump_sw=1;
    else if(strcmp(*argv,"--type=single")==0)        type_sw=0;
    else if(strcmp(*argv,"--type=fan-out")==0)       type_sw=1;
    else if(strcmp(*argv,"--type=fan-in")==0)        type_sw=2;
    else if(strcmp(*argv,"--type=ordered-pairs")==0) type_sw=3;
    else if(strcmp(*argv,"--type=all-pairs")==0)     type_sw=4;
    else if((Sstr=is_prefix("--span=",*argv))){
      char* endptr;
      S=strtol(Sstr,&endptr,10);
      if(*endptr!=0) error("number expected after `--span='");
      if(S<1) error("--span has to be positive, not %d", S);
      if(S>300) error("--span too large");
    }else error("unknown option %s", *argv);
    argv++; argc--;
  }
  
  if(help){
    fprintf(stderr,
"Standard Binary Preprocessor                           mmi_clustering v"VERSION"\n"
"Usage: sbp [--sliding] [--type=T] [--span=S] [--weight] [--dump] [infile.bin]\n"
"\n"
"Reads infile.bin (or stdin if infile is omitted) and writes result to\n"
"stdout. Input is supposed to be an interleaved pair of streams composed of\n"
"32bit binary numbers (l0,r0,l1,r1,l2,r2,...), where l's are the `left' words\n"
"and r's are `right' words in (l,r) context used by mmi_tree. Options affect\n"
"behavior in the following way:\n"
"\n"
"--sliding  If specified, the input (a,b,c,d...y,z) is internally modified \n"
"           into (a,b,b,c,c,d, .. y,y,z) before further processing begins.\n"
"           Therefore (l0,r0)=(a,b) and (l1,r1)=(b,c) and so on. This is\n"
"           useful when directly processing raw texts. Note that this is\n"
"           not the default.\n"
"--type=single   for each position p in the stream, emit single (l_p,r_p) pair\n"
"--type=fan-out  for each p, emit (l_p,r_p),(l_p,r_(p+1))...(l_p,r_(p+S-1))\n"
"--type=fan-in   for each p, emit (l_p,r_p),(l_(p+1),r_p)...(l_(p+S-1),r_p)\n"
"--type=ordered-pairs  emit all pairs (l_i,r_j) such that p<=i<j<p+S\n"
"--type=all-pairs      emit all pairs within interval [p,p+S-1]\n"
"       --type defaults to `single'.\n"
"--span=S   number S>=1, specifying the size of the window within which the\n"
"           output is generated in each step. Defaults to S=1.\n"
"--weight  Normally, each occurrence of (l,r) pair is counted only once. With\n"
"          this option it is counted w-times where w=S-d and d is a distance\n"
"          from l to r ((l0,r0) have 0 distance, (l1,r0) have distance 1 as\n"
"          well as (l0,r1)).\n"
"--dump  Normally, the output format is (l,r,count) for each word, \n"
"        which is expected by mmi_tree. This option changes it into (l,r)\n"
"        so that it could be fed back into show.\n");
    return 0;
  }
  if(type_sw==0) S=1;
  
  int in=0;
  if(*argv){
    in=open(*argv,O_RDONLY);
    if(in<0) error("cannot open file %s (%m)", *argv);
  }
  int bfr[2*((S-1) + BLEN) + 1];   //+1 is only used in sliding mode
  if(sliding_sw) tread(in,bfr,1);
  read_block(in, bfr, S-1);
  for(;;){
    int l=read_block(in, bfr+2*(S-1), BLEN);
    if(l==0) break;
    switch(type_sw){
    case 0: //single
      for(int k=0; k<l; k++) emit(bfr+2*k, bfr+2*k+1);
      break;
    case 1: //fan-out
      for(int k=0; k<l; k++) for(int q=k; q<k+S; q++) emit(bfr+2*k, bfr+2*q+1);
      break;
    case 2: //fan-in
      for(int k=0; k<l; k++) for(int q=k; q<k+S; q++) emit(bfr+2*q, bfr+2*k+1);
      break;
    case 3: //ordered-pairs
      for(int k=0; k<l; k++) for(int p=k; p<k+S; p++) for(int q=p; q<k+S; q++)
        emit(bfr+2*p, bfr+2*q+1);
      break;
    case 4: //all-pairs
      for(int k=0; k<l; k++) for(int p=k; p<k+S; p++) for(int q=k; q<k+S; q++)
        emit(bfr+2*p, bfr+2*q+1);
      break;
    }
    copy(bfr, bfr+2*l, 2*(S-1) + 1);
  }
  if(fclose(stdout)<0) error("cannot write (%m)");  
  return 0;  
}

