#define ERROR_MSG_PREFIX "show: "
#include "base.cc"

#define BL 16384     //size of word buffer
#define WINDOW 9     //max size of lookahead or lookbehind
#define MAX_BL 9     //max --blocks files
#define SCREEN_WIDTH (80-10)  //screen width (used by --human option)

char* format_str="w";
char* blfile[MAX_BL];
int* eqmap[MAX_BL];
FF*  affinity[MAX_BL];
int numeric_sw=0, human_sw=0, histogram_sw=0, tree_sw=0, blocks_sw=0, S=1;
int blfile_ct=0;
int w_back, w_forth;

void load_block_files(char** abc, int A)
{
  for(int i=0; i<blfile_ct; i++){
    alloc(affinity[i],A+1);
    eqmap[i]=load_eqmap(blfile[i], abc, affinity[i], A);
  }
}

void check_format(char* f, int blocks_ct, int& back, int& forth)
{
  back=forth=0;
  while(*f){
    int event=0;
    if(*f=='w') event=1;
    else if(*f=='b'){
      event=1; f++;
      if(*f>'9' || *f<'0') error("b-number expected in %s", f-1);
      int block=*f-'0';
      if(block>=blocks_ct) error("undefined b-number %d in %s", block, f-1);
    }
    f++;
    if(event){
      if(*f=='+' || *f=='-'){
        f++;
        if(!*f) error("premature end of format string (index expected)");
        int idx=*f-'0';
        if(idx>WINDOW) error("index %s larger than allowed WINDOW=%d", f-2, WINDOW);
        if(f[-1]=='-'){
          idx=-idx;
          if(idx<back) back=idx;
        }else 
          if(idx>forth) forth=idx;
        f++;
      }
    }
  }
}

void print_word(char* f, int* stream, char** abc, int &position)
{
  while(*f){
    int blcks=-1;
    if(*f=='b'){
      f++;
      blcks=*f-'0';
    }else if(*f!='w') {putchar(*f++); position++; continue;}
    f++;
    int idx=0;
    if(*f=='+' || *f=='-'){
      f++;
      idx=*f-'0';
      if(f[-1]=='-') idx=-idx;
      f++;
    }
    if(blcks==-1){
      if(numeric_sw) position+=printf("%d", stream[idx]);
      else position+=printf("%s", abc[stream[idx]]);
    }else
      position+=printf("%d", eqmap[blcks][stream[idx]]);
  }
  position++;
  if(!human_sw || position>SCREEN_WIDTH){
    position=0; 
    if(printf("\n")<0) error("cannot write (%m)");
  }else if(printf(" ")<0) error("cannot write (%m)");;
}

int main(int argc, char *argv[])
{
  int help=0;
  argv++; argc--;
  while(*argv && argv[0][0]=='-'){
    char* t;
    if(strcmp("--help",*argv)==0 || strcmp("-h",*argv)==0) help=1;
    else if((t=is_prefix("--blocks=",*argv))){
      if(blfile_ct>MAX_BL) error("too many `--blocks' options");
      blfile[blfile_ct++]=t;
    }
    else if((t=is_prefix("--format=", *argv))) format_str=t;
    else if(strcmp("--numeric",*argv)==0) numeric_sw=1;
    else if(strcmp("--human",*argv)==0) human_sw=1;
    else if(strcmp("--histogram",*argv)==0) histogram_sw=1;
    else if(strcmp("--tree",*argv)==0) tree_sw=1;
    else if(strcmp("--stree",*argv)==0) tree_sw=2;
    else if(strcmp("--ptree",*argv)==0) tree_sw=3;
    else if(strcmp("--blocks",*argv)==0) blocks_sw=1;
    else if(strcmp("--eqmap",*argv)==0) blocks_sw=2;
    else if((t=is_prefix("--step=",*argv))){
      char* endptr;
      S=strtol(t,&endptr,10);
      if(*endptr!=0) error("number expected after `--step='");
      if(S<1) error("--step has to be positive, not %d", S);
      if(S>30) error("--step too large");
    }else error("unknown option %s", *argv);
    argv++; argc--;
  }
  if(help){
    printf(
"Show                                                   mmi_clustering v"VERSION"\n"
"Usage: (1) show [OPTIONS] [ file.bin | bin_directory ]\n"
"    where OPTIONS are\n"
"           --blocks=blfile  file defining equivalence on words. Up to 10\n"
"                            different equivalences may be specified.\n"
"           --numeric        print decimal numbers rather than words\n"
"           --human          print in row rather than each word on separate line\n"
"           --format=string  template of output for single word, defaults to \"w\"\n"
"           --step=S         emit only each Sth word, defaults to S=1\n"
"       (2) show --histogram bin_directory\n"
"       (3) show [((--blocks=blfile)* [--tree|--stree]) | --ptree] file.tree\n"
"       (4) show --blocks=blkfile (--blocks | --eqmap) bin_directory\n"
"\n"
"The first form transform binary file back into its textual representation.\n"
"Input is read either from file or from stdin iff bin_directory is specified.\n"
"The actual shape of the output is controlled via a format string, obeying\n"
"following grammar:\n"
"\n"
" <format> = <fill>* ( (`w' | `b' <digit>)[( `+' | `-' ) <digit>] <fill>* )*\n"
" <digit>  = `0' | ... | `9'\n"
" <fill>   = any ASCII character except `w' and `b'\n"
"\n"
"Example: \"w b0_b2+3 w-3_b1\". Each output token produced is generated by\n"
"substituting word for any `w' clause, and number of equivalence block for\n"
"any b<digit> clause (digit selects equivalence file in order they were\n"
"specified by `--blocks' option), leaving <fill> characters intact. +-<digit>\n"
"clause controls where in the stream the actual token is taken from,\n"
"relatively to current position.\n"
"\n"
"The second form prints histogram of words in human readable form.\n"
"\n"
"The third form prints tree in human readable (--tree,--stree) or portable\n"
"machine readable form (--ptree).\n"
"\n"
"Last form reads single lock file and outputs it either as an eqmap-file or\n"
"a block-file (useful for sorting manually written block files).\n"
);
    return 0;
  }
  if(!*argv) error("file or directory name expected"); 
  char** abc;
  if(tree_sw){
    char dirname[2048]; mmi_result r; i64* hist;
    load_tree(*argv, r, dirname, abc);
    if(tree_sw<=2){
      printf("File %s using dictionary %s/dictionary\n", *argv, dirname);
      for(int k=0; k<blfile_ct; k++)
        printf("            BLCK%d=%s\n", k, blfile[k]);
      load_block_files(abc,r.A-1);
      load_histogram(hist, dirname);
      draw_tree(stdout,PRINT_LEVELS|PRINT_FREQS|(tree_sw==2?PRINT_STREE:0),r,abc,hist,eqmap,affinity,blfile_ct);
    }else
      print_portable_tree(stdout,r,abc);
  }else{
    char* bin_dir=*argv;
    if(!directory_exists(bin_dir)){
      bin_dir=strdup(*argv);
      char* last=strrchr(bin_dir,'/');
      if(!last) error("directory %s does not exist", bin_dir);
      *last=0;
      if(!directory_exists(bin_dir)) error("directory %s does not exist", bin_dir);
    }
    int A=load_dictionary(abc, bin_dir);    
    load_block_files(abc,A);
    if(histogram_sw){
      i64 *hist;
      int N=load_histogram(hist, bin_dir);
      for(int k=0; k<N; k++) printf("%6d %30s %10lld %7.5f\n", k, abc[k], hist[k], hist[k]/double(hist[0]));
    }else if(blocks_sw){
      if(blfile_ct==0) error("no block-file specified");
      if(blocks_sw==1) save_eqmap("/dev/stdout",abc,eqmap[0],affinity[0],A);
      else{
        for(int i=1; i<=A; i++)
          printf("%-40s %5d %10.8f\n", abc[i], eqmap[0][i], D(affinity[0][i]));
      }
    }else{
      check_format(format_str,blfile_ct, w_back, w_forth);
      int bfile=0;
      if(*argv!=bin_dir)
        if((bfile=open(*argv, O_RDONLY))<0) error("cannot open %s for reading (%m)", *argv);
      int overlap=w_forth-w_back;
      int bfr[BL+overlap];
      if(tread(bfile, bfr, overlap)<0) error("cannot read (%m)");
      int position=0, s=0;
      for(;;){
        int l=tread(bfile, bfr+overlap, BL);
        if(l<0) error("cannot read (%m)");
        if(l==0) break; 
        for(int k=0;k<l;k++)
          if(bfr[k]<1 || bfr[k]>A) error("invalid code %d (A=%d) found in the buffer", bfr[k],A);
        for(int k=0; k<l; k++,s++) if(s%S==0) print_word(format_str, bfr+k-w_back, abc, position);
        copy(bfr, bfr+l, overlap);
      }
      close(bfile);
    }
  }
  if(fclose(stdout)<0) error("cannot write (%m)");
  return 0;
}

