// >>>>> stp.cc

#define ERROR_MSG_PREFIX "stp: "
#include "cfg.h"
#include <os.cc>
#ifdef ALLOW_ESCII
#include <escii.h>
#endif

int sw_i=0, sw_w=0, sw_W=0, sw_c=0;
char* scan_format=" %1000s%*[^\n]\n";   //default
void bailout() {exit(1);}

static inline void edit(char* s)
{
  for(int k=strlen(s)-1; k>=0; k--) if(s[k]=='\n') s[k]=' ';
#ifdef ALLOW_ESCII
  if(sw_i) sto_lower(s);
#endif
}

int la=' ';   //lookahead character (or EOF)
//returns 1=ok 0=eof 
int get_word(FILE* in, char* symbol)
{
  if(scan_format){  //simple scanning method  
    if(fscanf(in,scan_format, symbol)!=1){
      if(!feof(in))
        fprintf(stderr,"stp: WARNING: fscanf failed feof=%d (last=%s)  \n", feof(in), symbol);
      else
        return 0;
    }
    return 1;
  }
#ifdef ALLOW_ESCII
  else  //more sophisticated scanning method
  {
    while(la>=0 && is_white(la)) la=fgetc(in); //eat white
    int ltr=is_letter(la), i=0;
    while(la>=0 && is_letter(la)==ltr && !is_white(la) && i<999){
      symbol[i++]=la;
      la=fgetc(in);
    }
    symbol[i]=0;
    return la>=0;
  }
#else 
  error("get_word(): should not get there!");
#endif
}

int main(int argc, char** argv)
{
  int help=0;

  argv++; argc--;  
  while(*argv && argv[0][0]=='-'){
    if(strcmp(*argv,"--help")==0 || strcmp(*argv,"-h")==0) help=1;
    else if(strcmp(*argv,"-w")==0) scan_format=" %1000s ";
    else if(strcmp(*argv,"-c")==0) scan_format="%c";  //read `BEWARE!' section below!!!
#ifdef ALLOW_ESCII
    else if(strcmp(*argv,"-W")==0) scan_format=0;
    else if(strcmp(*argv,"-i")==0) sw_i=1;
#endif
    else error("unknown option %s", *argv);     
    argv++; argc--;
  }
  
  if(help){
    fprintf(stderr,
"Standard Text Preprocessor                             mmi_clustering v"VERSION"\n"
           "Usage: stp "
#ifdef ALLOW_ESCII
  "[-i] "
#endif
  "[-w"
#ifdef ALLOW_ESCII
  "|-W"
#endif
  "|-c] [input_file]\n"
  "Reads input_file (or stdin) and emits words separated by `\\n's to stdout.\n"
  "Normally exactly one word is expected on each input line, possibly\n"
  "followed by a comment (separated from the word itself by a space).\n"
  "Options altering this behavior follows:\n"
  "   -w    reads words separated by whitespaces.\n"
  "   -c    considers individual characters as word-symbols\n"
  "         (note that it replaces \\n by a space).\n"
#ifdef ALLOW_ESCII
  "   -i    ignore-case mode (tolower).\n"
  "   -W    acts like -w but separates words on non-letter characters, too.\n"
  "         Therefore `str8est' will be treated as `str' `8' and `est'.\n"
#endif 
  "NOTE that there is a limit on maximum word length being 1000 characters.\n" 
    );
    return 0;
  }
  char symbol[1001]; symbol[1000]=0; 
                     symbol[1]=0; //BEWARE! since %c conversion does not 
                                  //add terminating-0 we add it here and 
                                  //use the fact that exactly one character
                                  //will be used throughout the life of the pgm
  FILE *in;
  if(!argv[0]) in=stdin;
  else         in=fopen(argv[0],"r");
  if(!in) error("cannot open file %s (%m)", argv[0]);
  while(!feof(in)){
    if(!get_word(in,symbol)) break;
    edit(symbol);
    if(printf("%s\n", symbol)<0) error("cannot write (%m)");
  }
  return 0;
}

