// >>>>> tab.cc

#ifndef TAB_CC
#define TAB_CC
 
#include "inc.h"
#define HT_LEN 348617

struct hnod{
  i64 data;
  int ct;
    
  hnod* next;
  hnod** prev;
  hnod* chain;  
  char name[1];
};


//----- global variables -----

hnod** htab;           //global table
hnod *hash_chain=0;    //its main chain

//----- functions -----

static void init_htab() //must be called before the first htab operation
{ 
  alloc(htab, HT_LEN);
  fill(htab, (hnod*)NULL, HT_LEN);
}

static inline u32 hashf(char* name)
{
  u32 h=0;
  while (*name) {h = *name+((h<<3) | (h>>29)); name++;}
  return h%HT_LEN;
}

static inline hnod* hash_qry(char* name) UNUSED;
static inline hnod* hash_qry(char* name)
{  
  hnod* p;
  for(p = htab[hashf(name)]; p && strcmp(p->name, name)!=0; p=p->next);
  return p;
}

static inline void hlist_ins(u32 h, hnod* q)
{
  q->next = htab[h];  q->prev = &htab[h];
  if(htab[h]) htab[h]->prev = &(q->next); 
  htab[h]=q;
}

static inline void chain_ins(hnod* q, hnod* &chain) UNUSED;
static inline void chain_ins(hnod* q, hnod* &chain) {hnod* t=chain; chain = q; q->chain=t;}

static hnod* hash_ins(char *name, hnod* &chain) UNUSED;
static hnod* hash_ins(char *name, hnod* &chain)
{
  hnod* p = (hnod*)malloc(sizeof(hnod)+strlen(name));
  if (!p) outofmem();
  strcpy(p->name, name);
  hlist_ins(hashf(name),p);
  chain_ins(p, chain);
  return p;
}
 
static hnod* hash_cins(char *name, hnod* &new_item) UNUSED;
static hnod* hash_cins(char *name, hnod* &new_item)
{
  hnod* p;
  new_item=0;
  u32 h = hashf(name);
  for(p = htab[h]; p && strcmp(p->name, name)!=0; p=p->next);
  if(p) return p;

  p = (hnod*)malloc(sizeof(hnod)+strlen(name));
  if (!p) outofmem();
  strcpy(p->name, name);
  hlist_ins(h,p);
  new_item=p; 
  return p;
}

static inline void hash_ins(hnod* q) UNUSED;
static inline void hash_ins(hnod* q)
{
  u32 h = hashf(q->name);
  hlist_ins(h, q);
}

static inline void done_htab() //only works correctly for single-chained htabs
{
  for(hnod* p=hash_chain; p;) {
    hnod* t=p->chain;
    free(p);
    p=t;
  }
  hash_chain=0;
  free(htab);
}


#endif
