// >>>>> text2bin.cc
#define ERROR_MSG_PREFIX "text2bin.cc: "

#include <errno.h>
#include <stdio.h>
#include <dirent.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <inc.h>
#include "tab.cc"
#include "base.cc"

i64 total=0;  //total number of words from all input files
char* tp=0;   //text preprocessor call string (pgmname and arguments)
#define MAX_CONV 1000000  //limit on number of files in source directory


static int get_word(FILE* in, char* symbol)
{
  if(fscanf(in," %1000s%*[^\n]\n", symbol)!=1){
    if(!feof(in)) error("fscanf() failed feof=%d (last=%s)  \n", feof(in), symbol);
    return 0;
  }
  return 1;
}
                            

/* Returns number of distict words across all files */
static int construct_dictionary(char* srcdir)
{
  int dict_entry_ct=0;
  DIR* d=opendir(srcdir);
  if(!d) error("can't open source directory %s (%s)", srcdir, strerror(errno));
  struct dirent *de;
  while( (de=readdir(d))!=0 ) {
    if(strcmp(".", de->d_name)==0 || strcmp("..", de->d_name)==0 || strchr(de->d_name,'.')==0) continue;
    char path[2*PATH_MAX];
    sprintf(path,"%s/%s", srcdir, de->d_name);
    norm_path(path);
    if(directory_exists(path)) continue;
    FILE* in=fopen_infile(tp,path);
    printf("pass-1: processing %-55s \r", path); fflush(stdout);
    char* symbol=path; //reusing space!!!
    while(!feof(in)){
      if(!get_word(in,symbol)) break;
      hnod* new_item=0;
      hnod* node=hash_cins(symbol, new_item);
      if(new_item){
        chain_ins(new_item, hash_chain); 
        new_item->ct=dict_entry_ct; dict_entry_ct++;
        if(dict_entry_ct==INT_MAX-1) error("Too many distict words, cannot continue");
        new_item->data=0;
      }
      node->data++;
      total++;
      if(total==I64_MAX) error("Text too long, cannot continue");
    }
    fclose(in); bury_zombie();
  }
  closedir(d);  
  return dict_entry_ct;
}

char* conv_list[MAX_CONV];
int current_conv=0;

static void translate_files(char* src_dir, char* dest_dir) 
{
  DIR* d=opendir(src_dir);
  if(!d) error("can't open source directory %s (%s)", src_dir, strerror(errno));
  struct dirent *de;
  while( (de=readdir(d))!=0 ) {
    if(strcmp(".", de->d_name)==0 || strcmp("..", de->d_name)==0  || strchr(de->d_name,'.')==0) continue;
    char path[2*PATH_MAX]; 
    sprintf(path,"%s/%s", src_dir, de->d_name);
    norm_path(path);
    if(directory_exists(path)) continue;
    if(!file_exists(path)) error("file %s does not exits", path);
    FILE *in=fopen_infile(tp,path);
    char* new_name=strrchr(path,'/')+1;
    current_conv++;
    if(current_conv==MAX_CONV) error("too many files (%d)", current_conv);
    char new_path[2*PATH_MAX];
    sprintf(new_path, "%s/%s.bin", dest_dir, new_name);
    conv_list[current_conv-1]=strdup(strrchr(new_path,'/')+1);
    
    FILE *out=mkdir_fopen(new_path, "w", 0644);
    if(!out) error("can't open output file %s", new_path);
    
    printf("pass-2: processing %-55s \r", path); fflush(stdout);

    char *symbol=path; //reusing allocated space !!   
    while(!feof(in)){
      if(!get_word(in,symbol)) break;
      hnod* q=hash_qry(symbol); 
      if (!q) error("internal error: word `%s' was not found in the table", symbol);
      put32(out, q->data);
    }
    fclose(in); fclose(out); bury_zombie();
  }
  closedir(d);
}

static int ccmp(const void*a, const void* b) { return strcmp(*(char**)a, *(char**)b); }
static inline int cmp(const void* a, const void* b) 
{ 
  int c=isaturate((*(hnod**)b)->data - (*(hnod**)a)->data); 
  if(!c) return (*(hnod**)a)->ct - (*(hnod**)b)->ct;
  else return c;
}

//DECLARE_hsort(hnod*, (cmp(&high, &low)))
/* It was found to be slower than qsort after all. On tested on text leading
 * to 176363 word dictionary, hsort took 249895s while qsort took only
 * 162942s.
 */


int main(int argc, char** argv)
{
  int err=0, help=0;
  argv++; argc--;
  if(argc==0) {err=help=1;}
  while(*argv && argv[0][0]=='-'){
    if(strcmp(*argv,"--help")==0 || strcmp(*argv,"-h")==0) help=1;
    else if(strcmp(*argv,"--tp")==0){
      argv++; argc--;
      int l=0;
      while(*argv && strcmp(*argv, "--")!=0){
        int m=strlen(*argv);
        resize(tp,l+1+m);
        copy(tp+l,*argv,m);
        l+=m+1;
        tp[l-1]=''; tp[l]=0;
        argv++; argc--;
      }
      if(l==0) error("preprocessor name expected");
      if(!*argv) error("unterminated preprocessor call (`--' expected)");
      tp[l-1]=0;
    }else error("unknown option %s", *argv);
    argv++; argc--;
  }  
  check_ps(tp);
  if(help){
    fprintf(stderr,
"Text to binary convertor                               mmi_clustering v"VERSION"\n"
"Usage: text2bin [--tp TP A1 ... An --] src_dir dest_dir\n"
"\n"
"Converts files from src_dir into binary format to be stored in the dest_dir\n"
"(which will be created if it does not already exist). All words of the input\n"
"files must be on separate lines (program treats each line as a `word'). If\n"
"the input does not conform to this, --tp option can be used to filter\n"
"files thru a preprocessor. Here TP is the name of the preprocessor (usually stp\n"
"-- Standard Text Preprocessor) and A1 to An are its arguments. Note that the\n"
"text is read (and run thru preprocessor) twice and that the preprocessor\n"
"MUST return identical results in both runs. Preprocessor has to read from\n"
"file specified by %%s symbol (up to 20 occurences (each representing\n"
"the same file) may appear between --tp and --) and write to stdout in\n"
"``one word per line'' format. Note that text2bin has word length limited to\n"
"1000 characters (overflown characters are ignored).\n"
"\n"
"Note that src_dir must be different from dest_dir. All regular files having\n"
"`.' in their name from src_dir are converted into the binary form and stored\n"
"in dest_dir with additional .bin extension. Binary format is a sequence of\n"
"32bit numbers, each being index into the dictionary (starting by 1) which is\n"
"created from all the processed text files and sorted according to word's\n"
"frequency. Dictionary (listing words in their order) will be stored in\n"
"dest_dir/dictionary. Binary form of the respective histogram will be\n"
"stored into dest_dir/histogram.\n"
"\n"
"Examples:  \n"
"Simple processing of text (words separated by ` ' and `\\n')\n"
"  text2bin --tp stp -w %%s -- ../in ../bin\n"
"A trick using bash to run data thru several filters also showing that\n"
"binary data can be reinterpreted again. Note that since files without `.'\n"
"in their name are not processed, it is ok to directly use ../bin containing\n"
"dictionary and histogram files.\n"
"  text2bin --tp bash -c \"show %%s | stp -w\" -- ../bin ../bin2\n"
    );
    return err;
  }  
  if(argc!=2) error("source and destination directories expected");
  if(same_file(argv[0],argv[1])) error("source directory must not be identical to destination directory");
  if(strlen(argv[0])>PATH_MAX-20) error("source directory name is too long");
  if(strlen(argv[1])>PATH_MAX-20) error("destination directory name is too long");
  
  init_htab();
  int n=construct_dictionary(argv[0]);  
  printf("\n");
  
  hnod** names;
  alloc(names, n);
  { int i=0; for(hnod* ch=hash_chain; ch; ch=ch->chain, i++) names[i]=ch; }
  qsort(names, n, sizeof(hnod*), cmp);
  //hsort(names, n);
  i64 histogram[n+1];
  histogram[0]=total;
  for(int i=0; i<n; i++)
    histogram[i+1]=names[i]->data;
  char alpha[PATH_MAX+1];
  sprintf(alpha, "%s/dictionary", argv[1]);
  FILE *alph=mkdir_fopen(alpha, "w", 0644);
  if(!alph) error("can't open file %s  (%s)", alpha, strerror(errno));
  sprintf(alpha, "%s/histogram", argv[1]);
  int fh=open(alpha, O_WRONLY|O_TRUNC|O_CREAT, 0644);
  if(fh<0) error("cannot create file %s (%m)", alpha);
  if(twrite(fh, histogram, n+1)!=n+1 || close(fh)<0)
    error("cannot write to file %s (%m)", alpha);
  printf("Dictionary:\n");
  for(int i=0; i<n; i++){
    if(i<20) printf("  %20s %10.5f\n", names[i]->name, names[i]->data/double(total));
    fprintf(alph, "%s\n", names[i]->name);
    names[i]->data=i+1;
  }
  printf("\n"); fclose(alph); free(names); 
  
  translate_files(argv[0], argv[1]);
  printf("\n");

  char convs[PATH_MAX+1];
  sprintf(convs, "%s/list", argv[1]);
  FILE *cf=mkdir_fopen(convs, "w", 0644);
  if(!cf) error("Can't open file %s (%s)\n", convs, strerror(errno));
  qsort(conv_list, current_conv, sizeof(char*), ccmp);
  for(int i=0; i<current_conv; i++)
    if(fprintf(cf, "%s\n", conv_list[i])<0) error("cannot write (%m)");
  if(fclose(cf)<0) error("cannot write (%m)");

  printf("done (%d files)\n", current_conv);
  return 0;
}


