// >>>>> unicode.cc

/* Some unicode (utf8) related functions
 */

#include <inc.h>

/* Decodes one character out of utf8 stream `a', and positions `a' to the
 * next character. In case of invalid code it returns 0xFFFF (unicode's
 * internal state-signalling character) to signal error. `a' is positioned
 * on the character that caused the error, unless it would not change. In
 * such a casae it points to the next character.
 */
u32 decode_utf8(char* &a)
{
  u8 xx=*a; a++;
  if(xx<0x80) return xx;  
  if(xx>=0x80 && xx<0xC2) return 0xFFFF;  //utf8-error
  
  u8 yy=xx; xx=*a; a++;
  if(yy<0xE0) {
    if(xx<0x80 || xx>0xBF) {a--; return 0xFFFF;}  //utf8-error
    return (xx&0x3F) | ((yy&0x1F)<<6);
  }
  u8 zz=yy; yy=xx; xx=*a; a++;
  if(zz<0xF0){
    if(! ( zz==0xE0 && 0xA0<=yy && yy<=0xBF ||
           0xE1<=zz && zz<=0xEC && 0x80<=yy && yy<=0xBF ||
           zz==0xED && 0x80<=yy && yy<=0x9F ||
           0xEE<=zz && zz<=0xEF && 0x80<=yy && yy<=0xBF ) ) {a-=2; return 0xFFFF;}  //utf8-error
    if(xx<0x80 || xx>0xBF) {a--; return 0xFFFF;}  //utf8-error
    return (xx&0x3F) | ((yy&0x3F)<<6) | ((zz&0xF)<<12);
  }
  u8 uu=zz; zz=yy; yy=xx; xx=*a; a++;
  if(uu<0xF8){
    if(!(uu==0xF0 && 0x90<=zz && zz<=0xBF || 
         uu<=0xF1 && uu<=0xF3 && 0x80<=zz && zz<=0xBF ||
         uu==0xF4 && 0x80<=zz && zz<=0x8F )) {a-=3; return 0xFFFF;}  //utf8-error
    if(yy<0x80 || yy>0xBF) {a-=2; return 0xFFFF;}  //utf8-error
    if(xx<0x80 || xx>0xBF) {a--; return 0xFFFF;}   //utf8-error
    return (xx&0x3F) | ((yy&0x3F)<<6) | ((zz&0x3F)<<12) | ((uu&0x7)<<18);
  }
  a-=3; return 0xFFFF; //utf8-error
}


/* This function returns number of characters of the given utf8 string s. It
 * interprets following unicode characters as having zero width (ignores
 * them in counting):
 *  
 *    0x200D   zero width joiner
 *    0x200C   zero width non joiner
 *    0x200B   zw space
 *    0x00AD   soft hyphen
 *    0xFEFF   zw no-break-space
 *  -----
 *    0x20D0 .. 0x20FF combining diacritical marks for symbols
 *    0x0300 .. 0x036F combining diacritical marks for text 
 *    0xFE20 .. 0xFE2F combining half marks
 */
int utf8_strlen(char* s)
{
  int ct=0;
  while(*s){
    u32 unic=decode_utf8(s);
    switch(unic){
    case 0xFFFF:
    case 0x200D: case 0x200C: case 0x200B: case 0x00AD: case 0xFEFF: 

    case 0x20D0: case 0x20D1: case 0x20D2: case 0x20D3: case 0x20D4: case
    0x20D5: case 0x20D6: case 0x20D7: case 0x20D8: case 0x20D9: case 0x20DA:
    case 0x20DB: case 0x20DC: case 0x20DD: case 0x20DE: case 0x20DF: case
    0x20E0: case 0x20E1: case 0x20E2: case 0x20E3: case 0x20E4: case 0x20E5:
    case 0x20E6: case 0x20E7: case 0x20E8: case 0x20E9: case 0x20EA: case
    0x20EB: case 0x20EC: case 0x20ED: case 0x20EE: case 0x20EF: case 0x20F0:
    case 0x20F1: case 0x20F2: case 0x20F3: case 0x20F4: case 0x20F5: case
    0x20F6: case 0x20F7: case 0x20F8: case 0x20F9: case 0x20FA: case 0x20FB:
    case 0x20FC: case 0x20FD: case 0x20FE: case 0x20FF:
  
    case 0x0300: case 0x0301: case 0x0302: case 0x0303: case 0x0304: case
    0x0305: case 0x0306: case 0x0307: case 0x0308: case 0x0309: case 0x030A:
    case 0x030B: case 0x030C: case 0x030D: case 0x030E: case 0x030F: case
    0x0310: case 0x0311: case 0x0312: case 0x0313: case 0x0314: case 0x0315:
    case 0x0316: case 0x0317: case 0x0318: case 0x0319: case 0x031A: case
    0x031B: case 0x031C: case 0x031D: case 0x031E: case 0x031F: case 0x0320:
    case 0x0321: case 0x0322: case 0x0323: case 0x0324: case 0x0325: case
    0x0326: case 0x0327: case 0x0328: case 0x0329: case 0x032A: case 0x032B:
    case 0x032C: case 0x032D: case 0x032E: case 0x032F: case 0x0330: case
    0x0331: case 0x0332: case 0x0333: case 0x0334: case 0x0335: case 0x0336:
    case 0x0337: case 0x0338: case 0x0339: case 0x033A: case 0x033B: case
    0x033C: case 0x033D: case 0x033E: case 0x033F: case 0x0340: case 0x0341:
    case 0x0342: case 0x0343: case 0x0344: case 0x0345: case 0x0346: case
    0x0347: case 0x0348: case 0x0349: case 0x034A: case 0x034B: case 0x034C:
    case 0x034D: case 0x034E: case 0x034F: case 0x0350: case 0x0351: case
    0x0352: case 0x0353: case 0x0354: case 0x0355: case 0x0356: case 0x0357:
    case 0x0358: case 0x0359: case 0x035A: case 0x035B: case 0x035C: case
    0x035D: case 0x035E: case 0x035F: case 0x0360: case 0x0361: case 0x0362:
    case 0x0363: case 0x0364: case 0x0365: case 0x0366: case 0x0367: case
    0x0368: case 0x0369: case 0x036A: case 0x036B: case 0x036C: case 0x036D:
    case 0x036E: case 0x036F:

    case 0xFE20: case 0xFE21: case 0xFE22: case 0xFE23: case 0xFE24: case
    0xFE25: case 0xFE26: case 0xFE27: case 0xFE28: case 0xFE29: case 0xFE2A:
    case 0xFE2B: case 0xFE2C: case 0xFE2D: case 0xFE2E: case 0xFE2F:

    default: ct++;
    }
  }
  return ct;
}

